#!/usr/bin/env python
from .build_amber import ImplicitSolvent, ExplicitSolvent
import gc
from glob import glob
import json
from MDAnalysis.lib.util import convert_aa_code
from openbabel import pybel
from openmm import *
from openmm.app import *
import os
from pathlib import Path
from pdbfixer import PDBFixer
from pdbfixer.pdbfixer import Sequence
from rdkit import Chem
import shutil
from typing import Dict, List, Union

PathLike = Union[str, Path]
OptPath = Union[str, Path, None]
Sequences = List[Sequence]

class LigandError(Exception):
    def __init__(self, message='This system contains ligands which we cannot model!'):
        self.message = message
        super().__init__(self.message)


class LigandBuilder:
    def __init__(self, path: PathLike, lig: str, lig_number: int=0, file_prefix: str=''):
        self.path = path
        self.lig = path / lig 
        self.ln = lig_number
        self.out_lig = path / f'{file_prefix}{lig.stem}'

    def parameterize_ligand(self) -> None:
        """
        Ensures consistent treatment of all ligand sdf files, generating
        GAFF2 parameters in the form of .frcmod and .lib files. Produces
        a mol2 file for coordinates and connectivity and ensures that
        antechamber did not fail. Hydrogens are added in rdkit which 
        generally does a good job of this.
        """
        ext = self.lig.suffix
        self.lig = self.lig.stem

        convert_to_gaff = f'antechamber -i {self.lig}_prep.mol2 -fi mol2 -o \
                {self.out_lig}.mol2 -fo mol2 -at gaff2 -c bcc -s 0 -pf y -rn LG{self.ln}'
        parmchk2 = f'parmchk2 -i {self.out_lig}.mol2 -f mol2 -o {self.out_lig}.frcmod'
        
        tleap_ligand = f"""source leaprc.gaff2
        LG{self.ln} = loadmol2 {self.out_lig}.mol2
        loadamberparams {self.out_lig}.frcmod
        saveoff LG{self.ln} {self.out_lig}.lib
        quit
        """
        
        if ext == '.sdf':
            self.process_sdf()
        else:
            self.process_pdb()

        self.convert_to_mol2()
        os.system(convert_to_gaff)
        try:
            self.move_antechamber_outputs()
            self.check_sqm()
            os.system(parmchk2)
            leap_file, leap_log = self.write_leap(tleap_ligand)
            os.system(f'tleap -f {leap_file} > {leap_log}')
        except FileNotFoundError:
            raise LigandError(f'Antechamber failed! {self.lig}')
    
    def process_sdf(self) -> None:
        """
        Add hydrogens in rdkit. Atom hybridization is taken from the
        input sdf file and if this is incorrect, hydrogens will be wrong
        too.
        """
        mol = Chem.SDMolSupplier(f'{self.lig}.sdf')[0]
        molH = Chem.AddHs(mol, addCoords=True)
        with Chem.SDWriter(f'{self.lig}_H.sdf') as w:
            w.write(molH)
        
    def process_pdb(self):
        mol = Chem.MolFromPDBFile(f'{self.lig}.pdb')
        molH = Chem.AddHs(mol, addCoords=True)
        with Chem.SDWriter(f'{self.lig}_H.sdf') as w:
            w.write(molH)

    def convert_to_mol2(self) -> None:
        mol = list(pybel.readfile('sdf', f'{self.lig}_H.sdf'))[0]
        mol.write('mol2', f'{self.lig}_prep.mol2', True)

    def move_antechamber_outputs(self) -> None:
        """
        Remove unneccessary outputs from antechamber. Keep the
        sqm.out file as proof that antechamber did not fail.
        """
        os.remove('sqm.in')
        os.remove('sqm.pdb')
        shutil.move('sqm.out', f'{self.lig}_sqm.out')
        
    def check_sqm(self) -> int:
        """
        Checks for evidence that antechamber calculations exited
        successfully. This is always on the second to last line,
        and if not present, indicates that we failed to produce
        sane parameters for this molecule. In that case, I wish
        you good luck.
        """
        line = open(f'{self.lig}_sqm.out').readlines()[-2]

        if 'Calculation Completed' not in line:
            # make sqm.in more tolerant
            # manually run sqm
            # if it still fails then we raise an error
            raise LigandError(f'SQM failed for ligand {self.lig}!')

    def write_leap(self, inp: str) -> str:
        """
        Writes out a tleap input file and returns the path
        to the file.
        """
        leap_file = f'{self.path}/tleap.in'
        leap_log = f'{self.path}/leap.log'
        with open(leap_file, 'w') as outfile:
            outfile.write(inp)
            
        return leap_file, leap_log


class PLINDERBuilder(ImplicitSolvent):
    """
    Builds complexes consisting of a biomolecule pdb and small molecule ligand.
    Runs antechamber workflow to generate gaff2 parameters.
    """
    def __init__(self, 
                 path: PathLike, 
                 system_id: str,
                 out: PathLike, 
                 **kwargs):
        super().__init__(path / system_id, 'receptor.pdb', out / system_id, 
                         protein=True, rna=True, dna=True, phos_protein=True,
                         mod_protein=True, **kwargs)
        self.system_id = system_id
        self.ffs.append('leaprc.gaff2')
        self.build_dir = self.out / 'build'
        self.ions = None
    
    def build(self) -> None:
        ligs = self.migrate_files()

        if not ligs:
            print(f'No ligands!\n\n{self.pdb}')
            raise LigandError

        self.ligs = self.ligand_handler(ligs)
        self.assemble_system()

    def ligand_handler(self, ligs: List[PathLike]) -> List[PathLike]:
        ligands = []
        for i, lig in enumerate(ligs):
            lig_builder = LigandBuilder(self.build_dir, lig, i)
            lig_builder.parameterize_ligand()
            ligands.append(os.path.basename(lig)[:-4])

        return ligands
        
    def migrate_files(self) -> List[str]:
        os.makedirs(str(self.build_dir), exist_ok=True)
        os.chdir(self.build_dir) # necessary for antechamber outputs

        # grab the sequence file to complete protein modeling
        shutil.copy(str(self.path / 'sequences.fasta'),
                    str(self.build_dir))
        self.fasta = str(self.build_dir / 'sequences.fasta')
        
        # fix and move pdb
        self.prep_protein()

        # move ligand(s)
        ligands = []
        lig_files = self.path / 'ligand_files'
        ligs = [Path(lig) for lig in glob(str(lig_files) + '/*.sdf')]
        for lig in ligs:
            shutil.copy(str(lig), 
                        str(self.build_dir))

            if self.check_ligand(lig):
                ligands.append(lig.name)


        # handle any potential ions
        if self.ions is not None:
            self.ffs.append('leaprc.water.tip3p')
            self.place_ions()

        return ligands

    def place_ions(self) -> None:
        """
        This is horrible and I apologize profusely if you find yourself
        having to go through the following. Good luck.
        """
        pdb_lines = open(self.pdb).readlines()[:-1]

        if 'END' in pdb_lines[-1]:
            if 'TER' in pdb_lines[-2]:
                ln = -3
            else:
                ln = -2
        elif 'TER' in pdb_lines[-1]:
            ln = -2
        else:
            ln = -1

        try:
            next_atom_num = int(pdb_lines[ln][6:12].strip()) + 1
            next_resid = int(pdb_lines[ln][22:26].strip()) + 1
        except ValueError:
            print(f'ERROR: {self.pdb}')
            raise LigandError
        
        for ion in self.ions:
            for atom in ion:
                # HERE BE DRAGONS
                ion_line = f'ATOM  {next_atom_num:>5}'

                if atom[0].lower() in ['na', 'k', 'cl']:
                    ionname = atom[0] + atom[1]
                    ion_line += f'{ionname:>4}  {ionname:<3} '
                else:
                    ionname = atom[0].upper()
                    ion_line += f'{ionname:>3}   {ionname:<3}'

                coords = ''.join([f'{x:>8.3f}' for x in atom[2:]])
                ion_line += f'{next_resid:>5}    {coords}  0.00  0.00\n'

                pdb_lines.append(ion_line)
                pdb_lines.append('TER\n')

                next_atom_num += 1
                next_resid += 1

        pdb_lines.append('END')
        
        with open(self.pdb, 'w') as f:
            f.write(''.join(pdb_lines))

    def assemble_system(self) -> None:
        """
        Slightly modified from the parent class, now we have to add
        the ligand parameters and assemble a complex rather than just
        placing a biomolecule in the water box.
        """
        tleap_complex = [f'source {ff}' for ff in self.ffs]
        structs = [f'PROT = loadpdb {self.pdb}']
        combine = 'COMPLEX = combine{PROT'
        for i, lig in enumerate(self.ligs):
            ligand = self.build_dir / lig
            tleap_complex += [f'loadamberparams {ligand}.frcmod', 
                              f'loadoff {ligand}.lib']
            structs += [f'LG{i} = loadmol2 {ligand}.mol2']
            combine += f' LG{i}'
        
        combine += '}'
        tleap_complex += structs
        tleap_complex.append(combine)
        tleap_complex += [
            'set default PBRadii mbondi3',
            f'savepdb COMPLEX {self.out}/system.pdb',
            f'saveamberparm COMPLEX {self.out}/system.prmtop {self.out}/system.inpcrd',
            'quit'
        ]

        tleap_complex = '\n'.join(tleap_complex)
        leap_file = self.build_dir / 'tleap.in'
        with open(str(leap_file), 'w') as outfile:
            outfile.write(tleap_complex)
            
        tleap = f'tleap -f {leap_file}'
        os.system(tleap)

    def prep_protein(self) -> None:
        raw_pdb = self.path / self.pdb
        prep_pdb = self.build_dir / 'prepped.pdb'
        self.pdb = self.build_dir / 'protein.pdb'
       
        # complex workflow for modeling missing residues
        self.triage_pdb(raw_pdb, prep_pdb)
        
        # remove hydrogens (-y) and waters (-d) from the input PDB
        pdb4amber = f'pdb4amber -i {prep_pdb} -o {self.pdb} -y -d'
        os.system(pdb4amber)

    def triage_pdb(self, broken_pdb: PathLike, 
                   repaired_pdb: PathLike) -> str:
        """
        Runs PDBFixer to repair missing loops and ensure structure is
        in good shape. Runs a check against the sequence provided by
        PLINDER and ensures that any non-canonical residues are represented
        in the sequence properly.
        """
        fixer = PDBFixer(filename=str(broken_pdb))
        chains = [chain for chain in fixer.topology.chains()]
        chain_map = {chain.id: [res for res in chain.residues()] 
                     for chain in chains}

        # non-databank models do not have SEQRES and therefore no
        # sequence data to model missing residues
        fixer.sequences = self.inject_fasta(chain_map)
        fixer.findMissingResidues()
        fixer.findMissingAtoms()
        fixer.addMissingAtoms()
        
        with open(str(repaired_pdb), 'w') as f:
            PDBFile.writeFile(fixer.topology, fixer.positions, f)

    def inject_fasta(self, 
                     chain_map: Dict[str, List[str]]) -> Sequences:
        """
        Checks fasta against actual sequence. Modifies sequence so that 
        it correctly matches in the case of non-canonical residues such
        as phosphorylations (i.e. SER -> SEP).
        """
        fasta = open(self.fasta).readlines()
        remapping = json.load(open(f'{self.path}/chain_mapping.json', 'rb'))
        sequences = []
        for i in range(len(fasta) // 2):
            seq_chain = fasta[2*i].strip()[1:] # strip off > and \n
            chain = remapping[seq_chain]
            one_letter_seq = fasta[2*i+1].strip()
            try:
                three_letter_seq = [convert_aa_code(aa) for aa in one_letter_seq]
            except ValueError:
                print(f'\nUnknown residue in fasta!\n\n{self.pdb}')
                raise LigandError

            try:
                three_letter_seq = self.check_ptms(three_letter_seq,
                                                   chain_map[chain])
                sequences.append(
                    Sequence(chainId=chain, 
                             residues=three_letter_seq)
                )
            except KeyError: # not sure what to do if this fails
                print(f'\nUnknown ligand error!\n\n{self.pdb}')
                raise LigandError

        return sequences

    def check_ptms(self, 
                   sequence: List[str],
                   chain_residues: List[str]) -> List[str]:
        """
        Check the full sequence (from fasta) against the potentially partial
        sequence from the structural model stored in `chain_residues`.
        """
        for residue in chain_residues:
            resID = int(residue.id) - 1 # since 0-indexed in list
            
            try:
                if sequence[resID] != residue.name:
                    sequence[resID] = residue.name
            except IndexError:
                print(f'Sequence length is messed up!\n\n{self.pdb}')
                raise LigandError
        
        return sequence

    def check_ligand(self, ligand: PathLike) -> bool:
        """
        Check ligand for ions and other weird stuff. We need to take care not
        to assume all species containing formal charges are ions, nor that all
        species containing atoms in the cation/anion lists are ions. Good example
        is the multitude of small molecule drugs containing bonded halogens.
        """
        ion = False
        mol = Chem.SDMolSupplier(str(ligand))[0]
        
        ligand = []
        for atom, position in zip(mol.GetAtoms(), mol.GetConformer().GetPositions()):
            symbol = atom.GetSymbol()
            if symbol.lower() in self.cation_list + self.anion_list:
                charge = atom.GetFormalCharge()
                if charge != 0:
                    ion = True
                    sign = '+' if charge > 0 else '-'
                    if abs(charge) > 1:
                        sign = f'{charge}{sign}'

                    ligand.append([symbol, sign] + [x for x in position])

        if ion:
            try:
                self.ions.append(ligand)
            except AttributeError:
                self.ions = [ligand]
            return False

        return True

    @property
    def cation_list(self) -> List[str]:
        return [
            'na', 'k', 'ca', 'mn', 'mg', 'li', 'rb', 'cs', 'cu',
            'ag', 'au', 'ti', 'be', 'sr', 'ba', 'ra', 'v', 'cr',
            'fe', 'co', 'zn', 'ni', 'pd', 'cd', 'sn', 'pt', 'hg',
            'pb', 'al'
        ]

    @property
    def anion_list(self) -> List[str]:
        return [
            'cl', 'br', 'i', 'f'
        ]
  
    
class ComplexBuilder(ExplicitSolvent):
    """
    Builds complexes consisting of a biomolecule pdb and small molecule ligand.
    Runs antechamber workflow to generate gaff2 parameters. Can optionally
    supply precomputed frcmod/lib files by their path + suffix in the 
    lig_param_prefix argument (e.g. /path/to/lig.mol2 or /path/to/lig)
    """
    def __init__(self, path: str, pdb: str, lig: str | list[str], padding: float=10., 
                 lig_param_prefix: str | None = None, **kwargs):
        super().__init__(path, pdb, padding)
        self.lig = Path(lig) if isinstance(lig, str) else [Path(l) for l in lig]
        self.ffs.append('leaprc.gaff2')
        self.build_dir = self.out / 'build'

        if lig_param_prefix is None:
            self.lig_param_prefix = lig_param_prefix
        else:
            prefix = Path(lig_param_prefix)
            self.lig_param_prefix = prefix.parent / prefix.stem

        for key, value in kwargs.items():
            setattr(self, key, value)
    
    def build(self) -> None:
        self.build_dir.mkdir(exist_ok=True, parents=True)
        os.chdir(self.build_dir) # necessary for antechamber outputs

        if self.lig_param_prefix is None:
            if isinstance(self.lig, list):
                lig_paths = []
                for i, lig in enumerate(self.lig):
                    lig_paths += self.process_ligand(lig, i)

                self.lig = lig_paths

            else:
                self.lig = self.process_ligand(self.lig)
        else:
            self.lig = self.lig_param_prefix

        if hasattr(self, ion):
            self.add_ion_to_pdb()

        self.prep_pdb()
        self.assemble_system()

    def process_ligand(self, lig: PathLike, prefix: int | None = None) -> PathLike:
        if lig.parent != self.build_dir:
            shutil.copy(lig, self.build_dir)
        
        if prefix is None:
            prefix = ''
            
        lig_builder = LigandBuilder(self.build_dir, lig, file_prefix=prefix)
        lig_builder.parameterize_ligand()

        return lig_builder.lig

    def add_ion_to_pdb(self) -> None:
        ion = [line for line in open(self.ion).readlines() 
               if any(['ATOM' in line, 'HETATM' in line])]
        pdb = [line for line in open(self.pdb).readlines()]
        
        out_pdb = []
        for line in pdb:
            if 'END' in line:
                out_pdb.append(ion)
                out_pdb.append(line)
            else:
                out_pdb.append(line)

        with open(self.pdb, 'w') as f:
            f.write('\n'.join(out_pdb))
        
    def assemble_system(self, dim, num_ions) -> None:
        """
        Slightly modified from the parent class, now we have to add
        the ligand parameters and assemble a complex rather than just
        placing a biomolecule in the water box.
        """
        tleap_ffs = '\n'.join([f'source {ff}' for ff in self.ffs])
        tleap_complex = [
            tleap_ffs,
            'source leaprc.gaff2',
        ]
        
        if not isinstance(self.lig, list):
            self.lig = [self.lig]
        
        LABELS = []
        for i, lig in enumerate(self.lig):
            tleap_complex += [
                f'loadamberparams {lig}.frcmod',
                f'loadoff {lig}.lib',
                f'LG{i} = loadmol2 {lig}.mol2',
            ]

            LABELS.append(f'LG{i}')

        LABELS.append(f'PROT')
        LABELS = ' '.join(LABELS)

        tleap_complex = [
            f'PROT = loadpdb {self.pdb}',
            f'COMPLEX = combine {{LABELS}}',
            'setbox COMPLEX centers',
            f'set COMPLEX box {{{dim} {dim} {dim}}}',
            f'solvatebox COMPLEX {self.water_box} {{0 0 0}}',
            'addions COMPLEX Na+ 0',
            'addions COMPLEX Cl- 0',
            f'addIonsRand COMPLEX Na+ {num_ions} Cl- {num_ions}',
            f'savepdb COMPLEX {self.out}.pdb',
            f'saveamberparm COMPLEX {self.out}.prmtop {self.out}.inpcrd'
        ]
        
        leap_file = self.write_leap('\n'.join(tleap_complex))
        tleap = f'tleap -f {leap_file}'
        os.system(tleap)
        



from collections import defaultdict
import json
import matplotlib.pyplot as plt
import MDAnalysis as mda
from MDAnalysis.analysis.distances import distance_array
from MDAnalysis.lib.util import convert_aa_code
import numpy as np
import pandas as pd
from pathlib import Path
import seaborn as sns
from typing import Callable, Dict, List, Tuple, Union

PathLike = Union[Path, str]
Results = Dict[str, Dict[str, float]]

class PPInteractions:
    """
    Code herein adapted from: 
        https://www.biorxiv.org/content/10.1101/2025.03.24.644990v1.full.pdf
    Takes an input topology file and trajectory file, and highlights relevant
    interactions between two selections. To this end we first compute the 
    covariance matrix between the two selections, filter out all interactions
    which occur too far apart (11Å for positive covariance, 13Å for negative
    covariance), and examines each based on a variety of distance and angle
    cutoffs defined in the literature.
    """
    def __init__(self, 
                 top: PathLike, 
                 traj: PathLike,
                 out: PathLike,
                 sel1: str='chainID A',
                 sel2: str='chainID B',
                 cov_cutoff: Tuple[float]=(11., 13.),
                 sb_cutoff: float=6.,
                 hbond_cutoff: float=3.5,
                 hbond_angle: float=30.,
                 hydrophobic_cutoff: float=8.,
                 plot: bool=True):
        self.u = mda.Universe(top, traj)
        self.n_frames = len(self.u.trajectory)
        self.out = out
        self.sel1 = sel1
        self.sel2 = sel2
        self.cov_cutoff = cov_cutoff
        self.sb = sb_cutoff
        self.hb_d = hbond_cutoff
        self.hb_a = hbond_angle * 180 / np.pi
        self.hydr = hydrophobic_cutoff
        self.plot = plot

    def run(self) -> None:
        """
        Main function that runs the workflow. Obtains a covariance matrix,
        screens for close interactions, evaluates each pairwise interaction
        for each amino acid and report the contact probability of each.
        """
        cov = self.get_covariance()
        positive, negative = self.interpret_covariance(cov)
        
        results = {'positive': {}, 'negative': {}}
        for res1, res2 in positive:
            data = self.compute_interactions(res1, res2)
            results['positive'].update(data)

        for res1, res2 in negative:
            data = self.compute_interactions(res1, res2)
            results['negative'].update(data)

        self.save(results)

        if self.plot:
            self.plot_results(results)

    def compute_interactions(self,
                             res1: int,
                             res2: int) -> Results:
        """
        Ingests two resIDs, generates MDAnalysis AtomGroups for each, identifies
        relevant non-bonded interactions (HBonds, saltbridge, hydrophobic) and
        computes each. Returns a Dict containing the proportion of simulation time
        that each interaction is engaged.
        """
        grp1 = self.u.select_atoms(f'{self.sel1} and resid {res1}')
        grp2 = self.u.select_atoms(f'{self.sel2} and resid {res2}')
        r1 = convert_aa_code(grp1.resnames[0])
        r2 = convert_aa_code(grp2.resnames[0])
        name = f'A_{r1}{res1}-B_{r2}{res2}'

        data = {name: {label: 0. for label in ['hydrophobic', 'hbond', 'saltbridge']}}
        function_calls, labels = self.identify_interaction_type(
            grp1.resnames[0], 
            grp2.resnames[0]
        )

        for call, label in zip(function_calls, labels):
            data[name][label] = call(grp1, grp2)

        return data

    def get_covariance(self) -> np.ndarray:
        """
        Loop over all C-alpha atoms and compute the positional
        covariance using the functional form:
            C = <(R1 - <R1>)(R2 - <R2>)T>
        where each element corresponds to the ensemble average movement
            C_ij = <deltaR_i * deltaR_j>
        with the magnitude being the strength of correlation and the sign
        corresponding to positive and negative correlation respectively.
        """
        p1_ca = self.u.select_atoms('chainID A and name CA')
        N = p1_ca.n_residues

        p2_ca = self.u.select_atoms('chainID B and name CA')
        M = p2_ca.n_residues

        self.res_map(p1_ca, p2_ca)

        R1_avg = np.zeros((N, 3))
        R2_avg = np.zeros((M, 3))

        for ts in self.u.trajectory:
            R1_avg += p1_ca.positions
            R2_avg += p2_ca.positions

        R1_avg /= self.n_frames
        R2_avg /= self.n_frames
        
        C = np.zeros((N, M))

        for ts in self.u.trajectory:
            R1 = p1_ca.positions
            R2 = p2_ca.positions

            dR1 = R1 - R1_avg
            dR2 = R2 - R2_avg

            for i in range(N):
                for j in range(M):
                    C[i, j] += np.dot(dR1[i], dR2[j])

        C /= self.n_frames
        
        for i in range(N):
            for j in range(M):
                dist = np.linalg.norm(R1_avg[i] - R2_avg[j])
                if C[i, j] > 0:
                    if dist > self.cov_cutoff[0]:
                        C[i, j] = 0.
                elif dist > self.cov_cutoff[1]:
                    C[i, j] = 0.

        return C

    def res_map(self,
                ag1: mda.AtomGroup,
                ag2: mda.AtomGroup) -> None:
        """
        Map covariance matrix indices to AtomGroup resIDs so that we are
        examining the correct pairs of residues.
        """
        mapping = {'ag1': {}, 'ag2': {}}
        for i, resid in enumerate(ag1.resids):
            mapping['ag1'][i] = resid

        for i, resid in enumerate(ag2.resids):
            mapping['ag2'][i] = resid

        self.mapping = mapping

    def interpret_covariance(self,
                             cov_mat: np.ndarray) -> Tuple[Tuple[int, int]]:
        """
        Identify pairs of residues with positive or negative correlations.
        Returns a tuple comprised of pairs for each.
        """
        pos_corr = np.where(cov_mat > 0.)
        neg_corr = np.where(cov_mat < 0.)
       
        seen = set()
        positive = list()
        for i in range(len(pos_corr[0])):
            res1 = self.mapping['ag1'][pos_corr[0][i]]
            res2 = self.mapping['ag2'][pos_corr[1][i]]
            if (res1, res2) not in seen:
                positive.append((res1, res2))
                seen.add((res1, res2))
                seen.add((res2, res1))

        negative = list()
        for i in range(len(neg_corr[0])):
            res1 = self.mapping['ag1'][neg_corr[0][i]]
            res2 = self.mapping['ag2'][neg_corr[1][i]]
            if (res1, res2) not in seen:
                negative.append((res1, res2))
                seen.add((res1, res2))
                seen.add((res2, res1))

        return positive, negative

    def identify_interaction_type(self,
                                  res1: str,
                                  res2: str) -> List[Callable]:
        """
        Identifies what analyses to compute for a given pair of protein
        residues (i.e. hydrophobic interactions, hydrogen bonds, saltbridges).
        """
        int_types = {
            'TYR': {'funcs': [self.analyze_hbond], 'label': ['hbond']}, 
            'HIS': {'funcs': [self.analyze_hbond], 'label': ['hbond']},
            'HID': {'funcs': [self.analyze_hbond], 'label': ['hbond']},
            'HIE': {'funcs': [self.analyze_hbond], 'label': ['hbond']},
            'SER': {'funcs': [self.analyze_hbond], 'label': ['hbond']},
            'THR': {'funcs': [self.analyze_hbond], 'label': ['hbond']},
            'ASN': {'funcs': [self.analyze_hbond], 'label': ['hbond']},
            'GLN': {'funcs': [self.analyze_hbond], 'label': ['hbond']},
            'ASP': {'funcs': [self.analyze_hbond, self.analyze_saltbridge],
                    'label': ['hbond', 'saltbridge']},
            'GLU': {'funcs': [self.analyze_hbond, self.analyze_saltbridge],
                    'label': ['hbond', 'saltbridge']},
            'LYS': {'funcs': [self.analyze_hbond, self.analyze_saltbridge],
                    'label': ['hbond', 'saltbridge']},
            'ARG': {'funcs': [self.analyze_hbond, self.analyze_saltbridge],
                    'label': ['hbond', 'saltbridge']},
            'HIP': {'funcs': [self.analyze_hbond, self.analyze_saltbridge], 
                    'label': ['hbond', 'saltbridge']},
        }
        
        funcs = defaultdict(lambda: [[], []])
        for res, calls in int_types.items():
            funcs[res] = [calls['funcs'], calls['label']]

        functions = [self.analyze_hydrophobic]
        labels = ['hydrophobic']
        for func, lab in zip(*funcs[res1]):
            if func in funcs[res2][0]:
                functions.append(func)
                labels.append(lab)

        return functions, labels

    def analyze_saltbridge(self,
                           res1: mda.AtomGroup,
                           res2: mda.AtomGroup) -> float:
        """
        Uses a simple distance cutoff to highlight the occupancy of 
        saltbridge between two residues. Returns the fraction of
        simulation time spent engaged in saltbridge.
        """
        pos = ['LYS', 'ARG']
        neg = ['ASP', 'GLU']
        name1 = res1.resnames[0]
        name2 = res2.resnames[0]
        if name1 not in pos + neg:
            return 0.
        elif name2 not in pos + neg:
            return 0.
        elif name1 in pos and name2 in pos:
            return 0.
        elif name1 in neg and name2 in neg:
            return 0.
        
        atom_names = ['NZ', 'NH1', 'NH2', 'OD1', 'OD2', 'OE1', 'OE2']

        grp1 = self.u.select_atoms('resname DUMMY')
        for atom in res1.atoms:
            if atom.name in atom_names:
                grp1 += atom

        grp2 = self.u.select_atoms('resname DUMMY')
        for atom in res2.atoms:
            if atom.name in atom_names:
                grp2 += atom
        
        n_frames = 0
        for ts in self.u.trajectory:
            dist = np.linalg.norm(grp1.positions - grp2.positions)
            if dist < self.sb:
                n_frames += 1
        
        return n_frames / self.n_frames

    def analyze_hbond(self,
                      res1: mda.AtomGroup,
                      res2: mda.AtomGroup) -> float:
        """
        Identifies all potential donor/acceptor atoms between two
        residues. Culls this list based on distance array across simulation
        and then evaluates each pair over the trajectory utilizing a
        distance and angle cutoff.
        """
        donors, acceptors = self.survey_donors_acceptors(res1, res2)

        n_frames = 0
        for ts in self.u.trajectory:
            n_frames += self.evaluate_hbond(donors, acceptors)

        return n_frames / self.n_frames

    def analyze_hydrophobic(self,
                            res1: mda.AtomGroup,
                            res2: mda.AtomGroup) -> float:
        h1 = self.u.select_atoms('resname DUMMY')
        h2 = self.u.select_atoms('resname DUMMY')

        for atom in res1.atoms:
            if 'C' in atom.type:
                h1 += atom

        for atom in res2.atoms:
            if 'C' in atom.type:
                h2 += atom

        n_frames = 0
        for ts in self.u.trajectory:
            da = distance_array(h1, h2)
            if np.min(da) < self.hydr:
                n_frames += 1

        return n_frames / self.n_frames

    def survey_donors_acceptors(self,
                                res1: mda.AtomGroup,
                                res2: mda.AtomGroup) -> Tuple[mda.AtomGroup]:
        """
        First pass distance threshhold to identify potential Hydrogen bonds.
        Should be followed by querying HBond angles but this serves to reduce
        our search space and time complexity. Only returns donors/acceptors which
        are within the distance cutoff in at least a single frame.
        """
        donors = self.u.select_atoms('resname DUMMY')
        acceptors = self.u.select_atoms('resname DUMMY')

        for atom in res1.atoms:
            if any([a in atom.type for a in ['O', 'N']]):
                if any(['H' in bond for bond in atom.bonded_atoms.types]):
                    donors += atom
                acceptors += atom
        
        for atom in res2.atoms:
            if any([a in atom.type for a in ['O', 'N']]):
                if any(['H' in bond for bond in atom.bonded_atoms.types]):
                    donors += atom
                acceptors += atom

        distances = distance_array(donors, acceptors)
        contacts = np.where(distances < self.hb_d)
        don_contacts = np.unique(contacts[0])
        acc_contacts = np.unique(contacts[1])

        return donors[don_contacts], acceptors[acc_contacts]

    def evaluate_hbond(sel,
                       donor: mda.AtomGroup,
                       acceptor: mda.AtomGroup) -> int:
        """
        Evaluates whether there is a defined hydrogen bond between any
        donor and acceptor atoms in a given frame. Must pass a distance
        cutoff as well as an angle cutoff. Returns early when a legal
        HBond is detected.
        """
        for d in donor.atoms:
            pos1 = d.position
            hpos = [atom.position for atom in d.bonded_atoms if 'H' in atom.type]
            for a in acceptor.atoms:
                pos3 = a.position

                if np.linalg.norm(pos3 - pos1) <= self.hb_d:
                    for pos2 in hpos:
                        v1 = pos2 - pos1
                        v2 = pos3 - pos2

                        v1 /= np.linalg.norm(v1)
                        v2 /= np.linalg.norm(v2)

                        if np.arccos(np.dot(v1, v2)) <= self.hb_a:
                            return 1

        return 0

    def save(self,
             results: Results) -> None:
        """
        Save results as a json file.
        """
        with open(self.out, 'w') as fout:
            json.dump(results, fout, indent=4)
    
    def plot_results(self,
                     results: Results) -> None:
        """
        """
        df = self.parse_results(results)
        
        plot = Path('plots')
        plot.mkdir(exist_ok=True)
        for cov_type in ['positive', 'negative']:
            for int_type in ['Hydrophobic', 'Hydrogen Bond', 'Salt Bridge']:
                data = df[(df['Covariance'] == cov_type) & (df[int_type] > 0.)]

                if not data.empty:
                    name = f'{cov_type.capitalize()}_Covariance_'
                    name += f'{"_".join(int_type.split(" "))}.png'

                    self.make_plot(
                        data,
                        int_type,
                        plot / name
                    )
    
    def parse_results(self, 
                      results: Results) -> pd.DataFrame:
        """
        Prepares results for plotting. Removes any entries which are
        all 0. and returns as a pandas DataFrame for easier plotting.
        """
        data_rows = []
        for cov_type, pair_dict in results.items():
            for pair, data in pair_dict.items():
                if any(val > 0. for val in data.values()):
                    row = {
                        'Residue Pair': pair,
                        'Hydrophobic': data['hydrophobic'],
                        'Hydrogen Bond': data['hbond'],
                        'Salt Bridge': data['saltbridge'],
                        'Covariance': cov_type,
                    }

                    data_rows.append(row)

        return pd.DataFrame(data_rows)
    
    def make_plot(self, 
                  data: pd.DataFrame,
                  column: str,
                  name: PathLike,
                  fs: int=15) -> None:
        fig, ax = plt.subplots(1, 1, figsize=(6, 5))

        sns.barplot(data=data, x='Residue Pair', y=column, ax=ax)
        
        ax.set_xlabel('Residue Pair', fontsize=fs)
        ax.set_ylabel('Probability', fontsize=fs)
        ax.set_title(column, fontsize=fs+2)
        ax.tick_params(labelsize=fs)
        ax.tick_params(axis='x', rotation=45)

        plt.tight_layout()
        plt.savefig(str(name), dpi=300)



#!/usr/bin/env python
from openmm.app import PDBFile
import os
from pathlib import Path
from pdbfixer import PDBFixer
import tempfile
from typing import Dict, List, Union

PathLike = Union[str, Path]
OptPath = Union[str, Path, None]

class ImplicitSolvent:
    """
    Class for building a system using ambertools. Produces explicit solvent cubic box
    with user-specified padding which has been neutralized and ionized with 150mM NaCl.
    """
    def __init__(self, path: OptPath, pdb: str, protein: bool=True,
                 rna: bool=False, dna: bool=False, phos_protein: bool=False,
                 mod_protein: bool=False, out: OptPath=None,
                 use_amber: bool=False, **kwargs):
        if path is None:
            self.path = Path(pdb).parent
        elif isinstance(path, str):
            self.path = Path(path)
        else:
            self.path = path

        self.path = self.path.resolve()
        self.path.mkdir(exist_ok=True, parents=True)

        self.pdb = pdb

        if out is not None:
            self.out = self.path / out
        else:
            self.out = self.path / 'protein.pdb' 

        self.out = self.out.resolve()

        self.use_amber = use_amber

        switches = [protein, rna, dna, phos_protein, mod_protein]
        ffs = [
            'leaprc.protein.ff19SB', 
            'leaprc.RNA.Shaw', 
            'leaprc.DNA.OL21',
            'leaprc.phosaa19SB',
            'leaprc.protein.ff14SB_modAA'
        ]
        
        self.ffs = [
            ff for ff, switch in zip(ffs, switches) if switch
        ]
        
        for key, val in kwargs.items():
            setattr(self, key, val)
        
        if not hasattr(self, 'tleap'):
            self.tleap = 'tleap'

    def build(self) -> None:
        """
        Orchestrate the various things that need to happen in order
        to produce an explicit solvent system. This includes running
        `pdb4amber`, computing the periodic box size, number of ions
        needed and running `tleap` to make the final system.
        """
        if self.use_amber:
            self.tleap_it()
        else:
            self.pdbfix_it()

    def pdbfix_it(self) -> str:
        """
        Using PDBFixer, add hydrogens and missing atoms and if a SEQ record
        exists (i.e. PDB is from RCSB) add missing residues. This is somewhat
        unstable so take care to make sure the outputs are ok.
        """
        fixer = PDBFixer(filename=str(self.path / self.pdb))
        fixer.findMissingResidues()
        fixer.findMissingAtoms()
        fixer.addMissingAtoms()
        fixer.addMissingHydrogens()
        
        with open(str(self.out), 'w') as f:
            PDBFile.writeFile(fixer.topology, fixer.positions, f)

    def tleap_it(self) -> None:
        """
        While more painful, tleap is more stable for system building.
        Runs the input PDB through tleap with the FF19SB protein forcefield
        and whichever other forcefields were turned on.
        """
        ffs = '\n'.join([f'source {ff}' for ff in self.ffs])
        tleap_in = f"""
        {ffs}
        prot = loadpdb {self.pdb}
        savepdb prot {self.out}
        saveamberparm prot {self.out.with_suffix('.prmtop')} {self.out.with_suffix('.inpcrd')}
        quit
        """

        self.temp_tleap(tleap_in)
    
    def write_leap(self, 
                   inp: str) -> Path:
        """
        Writes out a tleap input file and returns the path
        to the file.
        """
        leap_file = f'{self.path}/tleap.in'
        with open(leap_file, 'w') as outfile:
            outfile.write(inp)
            
        return Path(leap_file)

    def temp_tleap(self,
                   inp: str) -> None:
        """
        Writes a temporary file for tleap and then runs tleap. This
        makes handling parallel tleap runs much simpler as we are avoiding
        the different workers overwriting tleap inputs and getting confused.

        Arguments:
            inp (str): The tleap input file contents as a single string.
        """
        with tempfile.NamedTemporaryFile(mode='w+', suffix='.in') as temp_file:
            temp_file.write(inp)
            temp_file.flush()
            tleap_command = f'{self.tleap} -f {temp_file.name}'
            print(tleap_command)
            os.system(tleap_command)
    
class ExplicitSolvent(ImplicitSolvent):
    """
    Class for building a system using ambertools. Produces explicit solvent cubic box
    with user-specified padding which has been neutralized and ionized with 150mM NaCl.
    """
    def __init__(self, path: PathLike, pdb: PathLike, padding: float=10., protein: bool=True,
                 rna: bool=False, dna: bool=False, phos_protein: bool=False,
                 mod_protein: bool=False, polarizable: bool=False, **kwargs):
        super().__init__(path, pdb, protein, rna, dna, phos_protein, 
                         mod_protein, None, False, **kwargs)
        
        self.out = self.path / 'system'
        self.pad = padding
        self.ffs.extend(['leaprc.water.opc'])
        self.water_box = 'OPCBOX'
        
        if polarizable:
            self.ffs[0] = 'leaprc.protein.ff15ipq'
            self.ffs[-1] = 'leaprc.water.spceb'
            self.water_box = 'SPCBOX'
    
    def build(self) -> None:
        """
        Orchestrate the various things that need to happen in order
        to produce an explicit solvent system. This includes running
        `pdb4amber`, computing the periodic box size, number of ions
        needed and running `tleap` to make the final system.
        """
        self.prep_pdb()
        dim = self.get_pdb_extent()
        num_ions = self.get_ion_numbers(dim**3)
        self.assemble_system(dim, num_ions)
        self.clean_up_directory()

    def prep_pdb(self):
        """
        Runs input PDB through `pdb4amber` to ensure it is compliant enough
        that tleap won't freak out on us later. Removes any explicit hydrogens
        from the input structure to avoid name mismatches.
        """
        os.system(f'pdb4amber -i {self.pdb} -o {self.path}/protein.pdb -y')
        self.pdb = f'{self.path}/protein.pdb'
        
    def assemble_system(self, dim: float, num_ions: int) -> None:
        """
        Build system in tleap.
        """
        tleap_ffs = '\n'.join([f'source {ff}' for ff in self.ffs])
        tleap_complex = f"""{tleap_ffs}
        PROT = loadpdb {self.pdb}
        
        setbox PROT centers
        set PROT box {{{dim} {dim} {dim}}}
        solvatebox PROT {self.water_box} {{0 0 0}}
        
        addions PROT Na+ 0
        addions PROT Cl- 0
        
        addIonsRand PROT Na+ {num_ions} Cl- {num_ions}
        
        savepdb PROT {self.out}.pdb
        saveamberparm PROT {self.out}.prmtop {self.out}.inpcrd
        quit
        """
        
        leap_file = self.write_leap(tleap_complex)
        tleap = f'tleap -f {leap_file}'
        os.system(tleap) 

    def get_pdb_extent(self) -> int:
        """
        Identifies the longest axis of the protein in terms of X/Y/Z
        projection. Not super accurate but likely good enough for determining
        PBC box size. Returns longest axis length + 2 times the padding
        to account for +/- padding.
        """
        lines = [line for line in open(self.pdb).readlines() if 'ATOM' in line]
        xs, ys, zs = [], [], []
        
        for line in lines:
            x, y, z = [float(i.strip()) for i in line[26:54].split()]
            xs.append(x)
            ys.append(y)
            zs.append(z)
        
        xtent = (max(xs) - min(xs))
        ytent = (max(ys) - min(ys))
        ztent = (max(zs) - min(zs))
        
        return int(max([xtent, ytent, ztent]) + 2 * self.pad)
    
    def clean_up_directory(self) -> None:
        """
        Remove leap log. This is placed wherever the script calling it
        runs and likely will throw errors if multiple systems are
        being iteratively built.
        """
        os.remove('leap.log')
        (self.path / 'build').mkdir(exist_ok=True)
        for f in self.path.glob('*'):
            if not any([ext in f.name for ext in ['.prmtop', '.inpcrd', 'build']]):
                f.rename(f.parent / 'build' / f.name)
        
    @staticmethod
    def get_ion_numbers(volume: int) -> float:
        """
        Returns the number of Chloride? ions required to achieve 150mM
        concentration for a given volume. The number of Sodium counter
        ions should be equivalent.
        """
        return round(volume * 10e-6 * 9.03)



from .build_amber import ExplicitSolvent
import MDAnalysis as mda
import numpy as np
from pathlib import Path
from typing import Dict, List, Union
import yaml

PathLike = Union[Path, str]
Config = Dict[str, ...]

class InterfaceBuilder(ExplicitSolvent):
    def __init__(self, 
                 path: PathLike, 
                 pdb: str,
                 interfaces: Config,
                 target: PathLike,
                 binder: PathLike,
                 padding: float=10.,
                 protein: bool=True,
                 rna: bool=False,
                 dna: bool=False,
                 polarizable: bool=False):
        """
        For a given target/binder pair, build systems for driving binding to
        each of the supplied interfaces using DeepDriveMD. Includes writing
        out the required yaml files for running DeepDrive.
        """
        super().__init__(path, pdb, padding, protein, rna, dna, polarizable)
        self.interfaces = interfaces
        self.target = mda.Universe(target).select_atoms('all')
        self.binder = binder
        self.root = self.path / target.name[:-4] 
        self.com = self.target.center_of_mass()
    
    def build_all(self):
        """
        Iterates through each interface site for a given target and
        builds the corresponding system for the supplied miniprotein
        binder.
        """
        for site in self.interfaces.keys():
            # set pathing for this target/binder/site
            self.yaml_out = self.root / site / self.binder.name[:-4]
            self.out = self.yaml_out / 'ddmd' 
            self.out.mkdir(parents=True, exist_ok=True)
            self.out = self.out / 'system'
            self.build_dir = self.yaml_out / 'build'
            self.build_dir.mkdir(parents=True, exist_ok=True)
            self.pdb = self.build_dir / 'protein.pdb' # need full path to leverage parent methods

            cont_sel, dist_sel, vector, com, input_shape = self.parse_interface(site)

            binder = self.place_binder(np.array(vector, dtype=np.float32), 
                                       np.array(com, dtype=np.float32))
            self.merge_proteins(binder)

            self.path = self.build_dir
            self.build()
            
            self.write_ddmd_yaml(cont_sel, dist_sel)
            self.write_cvae_yaml(input_shape)

    def place_binder(self, 
                     vector: np.ndarray, 
                     com: np.ndarray) -> None:
        """
        Move binder nearby to the interface as defined by `vector`. Returns
        an MDAnalysis AtomGroup for the binder.
        """
        u = mda.Universe(self.binder)
        sel = u.select_atoms('all')
        binder_com = sel.center_of_mass()

        sel.positions -= com
        sel.positions += vector

        return sel

    def merge_proteins(self,
                       binder: mda.AtomGroup) -> None:
        """
        Merges the target and binder AtomGroups and writes out
        a unified PDB at `self.pdb` so as to leverage the existing
        pipeline for building explicit solvent systems.
        """
        merged_atoms = mda.Merge(self.target, binder)
        
        with mda.Writer(self.pdb) as W:
            W.write(merged_atoms)

    def parse_interface(self, 
                        site: str='site0') -> Config:
        """
        Returns the relevant data for the current interface site.
        """
        s = self.interfaces[site]
        N = len(s['contact_sel'][18:].split())
        inp_shape = (1, N, N)
        ret = [data for data in s.values()]
        ret.append(inp_shape)
        return ret # contact_sel, distance_sel, vector, com, inp_shape

    def write_ddmd_yaml(self,
                        contact_selection: str,
                        distance_selection: str) -> None:
        """
        Writes the simulation options yaml for a DeepDriveMD
        simulation.
        """
        yaml_settings = {
            'simulation_input_dir': 'ddmd',
            'num_workers': 4,
            'simulations_per_train': 6,
            'simulations_per_inference': 1,
            'num_total_simulations': 1000,
            'compute_settings': {
                'name': 'polaris',
                'num_nodes': 1,
                'worker_init': f'module use /soft/modulefiles; module load conda; \
                    conda activate deepdrive; cd {self.out}',
                'scheduler_options': '#PBS -l filesystems=home:eagle',
                'account': 'FoundEpidem',
                'queue': 'preemptable',
                'walltime': '72:00:00',
            },
            'simulation_settings': {
                'solvent_type': 'explicit',
                'dt_ps': 0.004,
                'mda_selection': contact_selection,
                'mda_selection_resid_list': None,
                'simulation_length_ns': 10,
                'report_interval_ps': 10,
                'temperature_kelvin': 300,
                'rmsd_reference_pdb': None,
                'distance_sels': distance_selection,
            },
            'train_settings': {
                'cvae_settings_yaml': 'cvae-prod-settings.yaml',
            },
        }

        with open(self.yaml_out / 'prod.yaml', 'w') as f:
            yaml.dump(yaml_settings, f)

    def write_cvae_yaml(self,
                        input_shape: List[int]) -> None:
        """
        Writes the CVAE options yaml for a DeepDriveMD
        simulation.
        """
        yaml_settings = {
            'input_shape': input_shape,
            'filters': [16, 16, 16, 16],
            'kernels': [3, 3, 3, 3],
            'strides': [1, 1, 1, 2],
            'affine_widths': [128],
            'affine_dropouts': [0.5],
            'latent_dim': 3,
            'lambda_rec': 1.0,
            'num_data_workers': 4,
            'prefetch_factor': 2,
            'batch_size': 64,
            'device': 'cuda',
            'optimizer_name': 'RMSprop',
            'optimizer_hparams': {
                'lr': 0.001,
                'weight_decay': 0.00001,
            },
            'epochs': 20,
            'checkpoint_log_every': 20,
            'plot_log_every': 20,
            'plot_n_samples': 5000,
            'plot_method': 'raw',
        }

        with open(self.yaml_out / 'cvae-prod-settings.yaml', 'w') as f:
            yaml.dump(yaml_settings, f)



from copy import deepcopy
from itertools import permutations
import numpy as np
from numpy import vectorize
from pathlib import Path
import polars as pl
from typing import Any, Dict, Union

PathLike = Union[Path, str]

class ipSAE:
    def __init__(self, 
                 structure_file: PathLike,
                 plddt_file: PathLike,
                 pae_file: PathLike,
                 out_path: PathLike | None):
        self.parser = ModelParser(structure_file)
        self.plddt_file = Path(plddt_file)
        self.pae_file = Path(pae_file)

        self.path = Path(out_path) if out_path is not None else self.plddt_file.parent
        self.path.mkdir(exist_ok=True)

    def parse_structure_file(self) -> None:
        self.parser.parse_structure_file()
        self.parser.classify_chains()
        self.coordinates = np.vstack([res['coor'] for res in self.parser.residues])
        self.token_array = np.array(self.parser.token_mask, dtype=bool)

    def prepare_scorer(self) -> None:
        chains = np.array(self.parser.chains)
        chain_types = self.parser.chain_types
        residue_types = np.array([res['res'] for res in self.parser.residues])

        self.scorer = ScoreCalculator(chains=chains,
                                      chain_pair_type=chain_types,
                                      residues=residue_types) 

    def run(self) -> None:
        self.parse_structure_file()

        distances = self.coordinates[:, np.newaxis, :] - self.coordinates[np.newaxis, :, :]
        distances = np.sqrt((distances ** 2).sum(axis=2))
        pLDDT = self.load_pLDDT_file()
        PAE = self.load_PAE_file()

        self.prepare_scorer()
        self.scorer.compute_scores(distances, pLDDT, PAE)

        self.scores = self.scorer.scores
        self.save_scores()

    def save_scores(self) -> None:
        self.scores.write_parquet(self.path / 'ipSAE_scores.parquet')

    def load_pLDDT_file(self) -> np.ndarray:
        data = np.load(str(self.plddt_file))
        pLDDT_arr = np.array(data['plddt'] * 100.)

        return pLDDT_arr

    def load_PAE_file(self) -> np.ndarray:
        data = np.load(str(self.pae_file))['pae']
        return data
    
        PAE = np.array(data['pae'])[np.ix_(self.token_array, self.token_array)]

        return PAE


class ScoreCalculator:
    def __init__(self,
                 chains: np.ndarray,
                 chain_pair_type: Dict[str, str],
                 residues: Any,
                 pdockq_cutoff: float=8.,
                 pae_cutoff: float=12.,
                 dist_cutoff: float=10.):
        self.chains = chains
        self.unique_chains = np.unique(chains)
        self.chain_pair_type = chain_pair_type
        self.residues = residues
        self.n_res = len(self.residues)
        self.pDockQ_cutoff = pdockq_cutoff
        self.PAE_cutoff = pae_cutoff
        self.dist_cutoff = dist_cutoff

        self.permute_chains()

    def compute_scores(self,
                       distances: np.ndarray,
                       pLDDT: np.ndarray,
                       PAE: np.ndarray) -> None:
        """
        Based on the input distance, pLDDT and PAE matrices, compute the pairwise pDockQ, pDockQ2,
        LIS, ipTM and ipSAE 
        """
        self.distances = distances
        self.pLDDT = pLDDT
        self.PAE = PAE

        results = []
        for chain1, chain2 in self.permuted:
            pDockQ, pDockQ2 = self.compute_pDockQ_scores(chain1, chain2)
            LIS = self.compute_LIS(chain1, chain2)
            ipTM, ipSAE = self.compute_ipTM_ipSAE(chain1, chain2)

            results.append([chain1, chain2, pDockQ, pDockQ2, LIS, ipTM, ipSAE])

        self.df = pl.DataFrame(np.array(results), schema={'chain1': str, 
                                                          'chain2': str, 
                                                          'pDockQ': float, 
                                                          'pDockQ2': float,
                                                          'LIS': float,
                                                          'ipTM': float,
                                                          'ipSAE': float})
        self.get_max_values()

    def compute_pDockQ_scores(self,
                              chain1: str,
                              chain2: str) -> tuple[float, float]:
        """
        Computes both the pDockQ and pDockQ2 scores for the interface between two chains.
        pDockQ is dependent solely on the pLDDT matrix while pDockQ2 is dependent on both
        pLDDT and the PAE matrix.

        Arguments:
            chain1 (str): The string name of the first chain.
            chain2 (str): The string name of the first chain.

        Returns:
            (tuple[float, float]): A tuple of the pDockQ and pDockQ2 scores respectively.
        """
        n_pairs = 0
        _sum = 0.
        residues = set()
        for i in range(self.n_res):
            if self.chains[i] == chain1:
                continue

            valid_pairs = (self.chains == chain2) & (self.distances[i] <= self.pDockQ_cutoff)
            n_pairs += np.sum(valid_pairs)
            if valid_pairs.any():
                residues.add(i)
                chain2_residues = np.where(valid_pairs)[0]
                pae_list = self.PAE[i][valid_pairs]
                pae_list_ptm = self.compute_pTM(pae_list, 10.)
                _sum += pae_list_ptm.sum()

                for residue in chain2_residues:
                    residues.add(residue)

        if n_pairs > 0:
            residues = list(residues)
            n_res = len(residues)
            mean_pLDDT = self.pLDDT[residues].mean()
            x = mean_pLDDT * np.log10(n_pairs)
            pDockQ = self.pDockQ_score(x)

            mean_pTM = _sum / n_pairs
            x = mean_pLDDT * mean_pTM
            pDockQ2 = self.pDockQ2_score(x)

        return pDockQ, pDockQ2

    def compute_LIS(self,
                    chain1: str, 
                    chain2: str) -> float:
        """
        Computes Local Interaction Score (LIS) which is based on a subset of the predicted aligned error 
        using a cutoff of 12. Values range in the interval (0, 1] and can be interpreted as how accurate
        a fold is within the error cutoff where a mean error of 0 yields a LIS value of 1 and a mean error
        that approaches 12 has a LIS value that approaches 0.

        Arguments:
            chain1 (str): The string name of the first chain.
            chain2 (str): The string name of the second chain.
        Returns:
            (float): The LIS value for both chains.
        """
        mask = (self.chains[:, None] == chain1) & (self.chains[None, :] == chain2)
        selected_pae = self.PAE[mask]

        LIS = 0.
        if selected_pae.size:
            valid_pae = selected_pae[selected_pae < 12]
            if valid_pae.size:
                scores = (12 - valid_pae) / 12
                avg_score = np.mean(scores)
                LIS = avg_score

        return LIS

    def compute_ipTM_ipSAE(self,
                           chain1: str,
                           chain2: str) -> tuple[float, float]:
        pair_type = 'protein'
        if 'nucleic' in [self.chain_pair_type[chain1], self.chain_pair_type[chain2]]:
            pair_type = 'nucleic'

        L = np.sum(self.chains == chain1) + np.sum(self.chains == chain2)
        d0_chain = self.compute_d0(L, pair_type)

        pTM_matrix_chain = self.compute_pTM(self.PAE, d0_chain)
        ipTM_byres = np.zeros((pTM_matrix_chain.shape[0]))

        valid_pairs_ipTM = (self.chains == chain2)
        ipTM_byres = np.array([0.])
        if valid_pairs_ipTM.any():
            ipTM_byres = np.mean(pTM_matrix_chain[:, valid_pairs_ipTM], axis=0)

        valid_pairs_matrix = (self.chains == chain2) & (self.PAE < self.PAE_cutoff)
        valid_pairs_ipSAE = valid_pairs_matrix

        ipSAE_byres = np.array([0.])
        if valid_pairs_ipSAE.any():
            ipSAE_byres = np.mean(pTM_matrix_chain[valid_pairs_ipSAE], axis=0)

        ipTM = np.max(ipTM_byres)
        ipSAE = np.max(ipSAE_byres)

        return ipTM, ipSAE

    def get_max_values(self) -> None:
        rows = []
        processed = set()
        for chain1, chain2 in self.permuted:
            if not all([chain in processed for chain in (chain1, chain2)]):
                filtered = self.df.filter(
                    ((pl.col('chain1') == chain1) & (pl.col('chain2') == chain2)) |
                    ((pl.col('chain1') == chain2) & (pl.col('chain2') == chain1))
                )
                max_ipsae = filtered.select('ipSAE').max().item()
                max_row = filtered.filter(pl.col('ipSAE') == max_ipsae)
                rows.append(max_row)

                processed.add(chain1)
                processed.add(chain2)

        self.scores = pl.concat(rows)

    def permute_chains(self) -> None:
        permuted = set()
        for c1, c2 in permutations(self.unique_chains):
            if c1 != c2:
                permuted.add((c1, c2))
                permuted.add((c2, c1))

        self.permuted = list(permuted)

    @staticmethod
    def pDockQ_score(x) -> float:
        return 0.724 / (1 + np.exp(-0.052 * (x - 152.611))) + 0.018

    @staticmethod
    def pDockQ2_score(x) -> float:
        return 1.31 / (1 + np.exp(-0.075 * (x - 84.733))) + 0.005

    @staticmethod
    @vectorize
    def compute_pTM(x: float,
                    d0: float) -> float:
        return 1. / (1 + (x / d0) ** 2)

    @staticmethod
    def compute_d0(L: int,
                   pair_type: str) -> float:
        L = max(27, L)

        min_value = 1.
        if pair_type == 'nucleic_acid':
            min_value = 2.

        return max(min_value, 1.24 * (L - 15) ** (1/3) - 1.8)


class ModelParser:
    def __init__(self, 
                 pdb: PathLike):
        self.pdb = pdb if isinstance(pdb, Path) else Path(pdb)

        self.token_mask = []
        self.residues = []
        self.cb_residues = []
        self.chains = []

    def parse_structure_file(self) -> None:
        if self.pdb.suffix == '.pdb':
            line_parser = self.parse_pdb_line
        else:
            line_parser = self.parse_cif_line

        field_num = 0
        lines = open(self.pdb).readlines()
        fields = dict()
        for line in lines:
            if line.startswith('_atom_site.'):
                _, field_name = line.strip().split('.')
                fields[field_name] = field_num
                field_num += 1

            if any([line.startswith(atom) for atom in ['ATOM', 'HETATM']]):
                atom = line_parser(line, fields)

                name = atom['atom_name']
                if name == 'CA':
                    self.token_mask.append(1)
                    self.residues.append(atom)
                    self.chains.append(atom['chain_id'])
                    if atom['res'] == 'GLY':
                        self.cb_residues.append(atom)

                elif 'C1' in name:
                    self.token_mask.append(1)
                    self.residues.append(atom)
                    self.chains.append(atom['chain_id'])

                elif name == 'CB' or 'C3' in name:
                    self.cb_residues.append(atom)

    def classify_chains(self) -> None:
        self.residue_types = np.array([res['res'] for res in self.residues])
        chains = np.unique(self.chains)
        self.chain_types = {chain: 'protein' for chain in chains}
        for chain in chains:
            indices = np.where(chains == chain)[0]
            chain_residues = self.residue_types[indices]
            if any([r in chain_residues for r in self.nucleic_acids]):
                self.chain_types[chain] = 'nucleic_acid'

    @property
    def nucleic_acids(self) -> list[str]:
        return ['DA', 'DC', 'DT', 'DG', 'A', 'C', 'U', 'G']

    @staticmethod
    def parse_pdb_line(line: str, *args) -> dict[str, Any]:
        atom_num = line[6:11].strip()
        atom_name = line[12:16].strip()
        residue_name = line[17:20].strip()
        chain_id = line[21]
        residue_id = line[22:26].strip()
        x = line[30:38].strip()
        y = line[38:46].strip()
        z = line[46:54].strip()

        return ModelParser.package_line(atom_num, atom_name, residue_name, chain_id, residue_id, x, y, z)

    @staticmethod
    def parse_cif_line(line: str, fields: dict[str, int]) -> dict[str, Any]:
        _split = line.split()
        atom_num = _split[fields['id']]
        atom_name = _split[fields['label_atom_id']]
        residue_name = _split[fields['label_comp_id']]
        chain_id = _split[fields['label_asym_id']]
        residue_id = _split[fields['label_seq_id']]
        x = _split[fields['Cartn_x']]
        y = _split[fields['Cartn_y']]
        z = _split[fields['Cartn_z']]

        if residue_id == '.':
            return None

        return ModelParser.package_line(atom_num, atom_name, residue_name, chain_id, residue_id, x, y, z)

    @staticmethod
    def package_line(atom_num: str,
                     atom_name: str,
                     residue_name: str,
                     chain_id: str,
                     residue_id: str,
                     x: str,
                     y: str,
                     z: str) -> dict[str, Any]:
        return {
            'atom_num': int(atom_num),
            'atom_name': atom_name,
            'coor': np.array([float(i) for i in [x, y, z]]),
            'res': residue_name,
            'chain_id': chain_id,
            'resid': int(residue_id),
        }



from copy import deepcopy
import MDAnalysis as mda
import numpy as np
from openmm import *
from openmm.app import *
from openmm.unit import *
from openmm.app.internal.singleton import Singleton
import os
from pathlib import Path
from typing import List, Tuple, Union

PathLike = Union[Path, str]

class Simulator:
    def __init__(self, 
                 path: PathLike, 
                 equil_steps: int=1_250_000, 
                 prod_steps: int=250_000_000, 
                 n_equil_cycles: int=3,
                 reporter_frequency: int=1_000,
                 platform: str='CUDA',
                 device_ids: list[int]=[0],
                 force_constant: float=10.):
        self.path = path if isinstance(path, Path) else Path(path)
        # input files
        self.prmtop = path / 'system.prmtop'
        self.inpcrd = path / 'system.inpcrd'

        # logging/checkpointing stuff
        self.eq_state = path / 'eq.state'
        self.eq_chkpt = path / 'eq.chk'
        self.eq_log = path / 'eq.log'
        self.eq_freq = reporter_frequency
        
        self.dcd = path / 'prod.dcd'
        self.restart = path / 'prod.rst.chk'
        self.state = path / 'prod.state'
        self.chkpt = path / 'prod.chk'
        self.prod_log = path / 'prod.log'
        self.prod_freq = self.eq_freq * 10

        # simulation details
        self.indices = self.get_restraint_indices()
        self.equil_cycles = n_equil_cycles
        self.equil_steps = equil_steps
        self.prod_steps = prod_steps
        self.k = force_constant
        self.platform = Platform.getPlatformByName(platform)
        self.properties = {'DeviceIndex': ','.join([str(x) for x in device_ids]), 
                           'Precision': 'mixed'}

        if platform == 'CPU':
            self.properties = {}

    def load_amber_files(self) -> System:
        if isinstance(self.inpcrd, Path | str):
            self.inpcrd = AmberInpcrdFile(str(self.inpcrd))
            self.prmtop = AmberPrmtopFile(str(self.prmtop), periodicBoxVectors=self.inpcrd.boxVectors)

        system = self.prmtop.createSystem(nonbondedMethod=PME,
                                          removeCMMotion=False,
                                          nonbondedCutoff=1. * nanometer,
                                          constraints=HBonds,
                                          hydrogenMass=1.5 * amu)
    
        return system
    
    def setup_sim(self, system: System, dt: float) -> Tuple[Simulation, Integrator]:
        integrator = LangevinMiddleIntegrator(300*kelvin, 
                                              1/picosecond, 
                                              dt*picoseconds)
        simulation = Simulation(self.prmtop.topology, 
                                system, 
                                integrator, 
                                self.platform, 
                                self.properties)
    
        return simulation, integrator

    def run(self) -> None:
        skip_eq = all([f.exists() 
                       for f in [self.eq_state, self.eq_chkpt, self.eq_log]])
        if not skip_eq:
            self.equilibrate()

        if self.restart.exists():
            self.check_num_steps_left()
            self.production(chkpt=self.restart, 
                            restart=True)
        else:
            self.production(chkpt=self.eq_chkpt,
                            restart=False)

    def equilibrate(self) -> Simulation:
        system = self.add_backbone_posres(self.load_amber_files(), 
                                          self.inpcrd.positions, 
                                          self.prmtop.topology.atoms(), 
                                          self.indices,
                                          self.k)
    
        simulation, integrator = self.setup_sim(system, dt=0.002)
        
        simulation.context.setPositions(self.inpcrd.positions)
        simulation.minimizeEnergy()
        
        simulation.reporters.append(StateDataReporter(str(self.eq_log), 
                                                      str(self.eq_freq), 
                                                      step=True,
                                                      potentialEnergy=True,
                                                      speed=True,
                                                      temperature=True))
        simulation.reporters.append(DCDReporter(str(self.path / 'eq.dcd'), self.eq_freq))

        simulation, integrator = self._heating(simulation, integrator)
        simulation = self._equilibrate(simulation)
        
        return simulation

    def production(self, chkpt: str, restart: bool=False) -> None:
        system = self.load_amber_files()
        simulation, integrator = self.setup_sim(system, dt=0.004)
        
        system.addForce(MonteCarloBarostat(1*atmosphere, 300*kelvin))
        simulation.context.reinitialize(True)

        if restart:
            log_file = open(str(self.prod_log), 'a')
        else:
            log_file = str(self.prod_log)

        simulation = self.load_checkpoint(simulation, chkpt)
        simulation = self.attach_reporters(simulation,
                                           self.dcd,
                                           log_file,
                                           self.restart,
                                           restart=restart)
    
        self._production(simulation)
    
    def load_checkpoint(self, 
                        simulation: Simulation, 
                        checkpoint: str) -> Simulation:
        simulation.loadCheckpoint(checkpoint)
        state = simulation.context.getState(getVelocities=True, getPositions=True)
        positions = state.getPositions()
        velocities = state.getVelocities()
        
        simulation.context.setPositions(positions)
        simulation.context.setVelocities(velocities)

        return simulation

    def attach_reporters(self, 
                         simulation: Simulation, 
                         dcd_file: str, 
                         log_file: str, 
                         rst_file: str, 
                         restart: bool=False) -> Simulation:
        simulation.reporters.extend([
            DCDReporter(
                dcd_file, 
                self.prod_freq,
                append=restart
                ),
            StateDataReporter(
                log_file,
                self.prod_freq,
                step=True,
                potentialEnergy=True,
                temperature=True,
                progress=True,
                remainingTime=True,
                speed=True,
                volume=True,
                totalSteps=self.prod_steps,
                separator='\t'
                ),
            CheckpointReporter(
                rst_file,
                self.prod_freq * 10
                )
            ])

        return simulation

    def _heating(self, 
                 simulation: Simulation, 
                 integrator: Integrator) -> Tuple[Simulation, Integrator]:
        simulation.context.setVelocitiesToTemperature(5*kelvin)
        T = 5
        
        integrator.setTemperature(T * kelvin)
        mdsteps = 100000
        length = mdsteps // 1000
        tstep = (300 - T) / length
        for i in range(length):
          simulation.step(mdsteps//60)
          temp = T + tstep * (1 + i)
          
          if temp > 300:
            temp = 300
          
          integrator.setTemperature(temp * kelvin)
    
        return simulation, integrator
         
    def _equilibrate(self, simulation: Simulation) -> Simulation:
        simulation.context.reinitialize(True)
        n_levels = 5
        d_k = self.k / n_levels
        eq_steps = self.equil_steps // n_levels

        for i in range(n_levels): 
            simulation.step(eq_steps)
            k = float(self.k - (i * d_k))
            simulation.context.setParameter('k', (k * kilocalories_per_mole/angstroms**2))
        
        simulation.context.setParameter('k', 0)
        simulation.step(eq_steps)
    
        simulation.system.addForce(MonteCarloBarostat(1*atmosphere, 300*kelvin))
        simulation.step(self.equil_cycles * eq_steps)

        simulation.system.addForce(MonteCarloBarostat(1*atmosphere, 300*kelvin))
        simulation.step(3*eq_steps)

        simulation.saveState(str(self.eq_state))
        simulation.saveCheckpoint(str(self.eq_chkpt))
    
        return simulation
    
    def _production(self, simulation: Simulation) -> Simulation:
        simulation.step(self.prod_steps)
        simulation.saveState(str(self.state))
        simulation.saveCheckpoint(str(self.chkpt))
    
        return simulation

    def get_restraint_indices(self, addtl_selection: str='') -> List[int]:
        u = mda.Universe(str(self.prmtop), str(self.inpcrd))
        if addtl_selection:
            sel = u.select_atoms(f'backbone or nucleicbackbone or {addtl_selection}')
        else:
            sel = u.select_atoms('backbone or nucleicbackbone')
            
        return sel.atoms.ix
        
    def check_num_steps_left(self) -> None:
        prod_log = open(str(self.prod_log)).readlines()

        try:
            last_line = prod_log[-1]
            last_step = int(last_line.split()[1].strip())
        except IndexError:
            try:
                last_line = prod_log[-2]
                last_step = int(last_line.split()[1].strip())
            except IndexError: # something weird happend just run full time
                return
        
        if time_left := (self.prod_steps - last_step):
            self.prod_steps -= time_left

            if n_repeat_timesteps := (last_step % (self.prod_freq * 10)):
                self.prod_steps -= n_repeat_timesteps
                n_repeat_frames = n_repeat_timesteps / self.prod_freq
                
                n_total_frames = last_step / self.prod_freq
                
                lines = [f'{n_total_frames - n_repeat_frames},{n_total_frames}']
                duplicate_log = self.path / 'duplicate_frames.log'
                if duplicate_log.exists():
                    mode = 'a'
                else:
                    mode = 'w'
                    lines = ['first_frame,last_frame'] + lines
                    
                with open(str(duplicate_log), mode) as fout:
                    fout.write('\n'.join(lines))

    @staticmethod
    def add_backbone_posres(system: System, positions: np.ndarray, atoms: List[str], 
                            indices: List[int], restraint_force: float=10.) -> System:
        force = CustomExternalForce("k*periodicdistance(x, y, z, x0, y0, z0)^2")
    
        force_amount = restraint_force * kilocalories_per_mole/angstroms**2
        force.addGlobalParameter("k", force_amount)
        force.addPerParticleParameter("x0")
        force.addPerParticleParameter("y0")
        force.addPerParticleParameter("z0")
    
        for i, (atom_crd, atom) in enumerate(zip(positions, atoms)):
            if atom.index in indices:
                force.addParticle(i, atom_crd.value_in_unit(nanometers))
      
        posres_sys = deepcopy(system)
        posres_sys.addForce(force)
      
        return posres_sys

class ImplicitSimulator(Simulator):
    def __init__(self, 
                 path: str, 
                 equil_steps: int=1_250_000, 
                 prod_steps: int=250_000_000, 
                 n_equil_cycles: int=3,
                 reporter_frequency: int=1_000,
                 platform: str='CUDA',
                 device_ids: list[int]=[0],
                 force_constant: float=10.,
                 implicit_solvent: Singleton=GBn2,
                 solute_dielectric: float=1.,
                 solvent_dielectric: float=78.5):
        super().__init__(path, equil_steps, prod_steps, n_equil_cycles,
                         reporter_frequency, platform, device_ids, 
                         force_constant)
        self.solvent = implicit_solvent
        self.solute_dielectric = solute_dielectric
        self.solvent_dielectric = solvent_dielectric
        # k = 367.434915 * sqrt(conc. [M] / (solvent_dielectric * T))
        self.kappa = 367.434915 * np.sqrt(.15 / (solvent_dielectric * 300))
    
    def load_amber_files(self) -> System:
        if isinstance(self.inpcrd, str):
            self.inpcrd = AmberInpcrdFile(self.inpcrd)
            self.prmtop = AmberPrmtopFile(self.prmtop)

        system = self.prmtop.createSystem(nonbondedMethod=NoCutoff,
                                          removeCMMotion=False,
                                          constraints=HBonds,
                                          hydrogenMass=1.5 * amu,
                                          implicitSolvent=self.solvent,
                                          soluteDielectric=self.solute_dielectric,
                                          solventDielectric=self.solvent_dielectric,
                                          implicitSolventKappa=self.kappa/nanometer)
    
        return system
        
    def equilibrate(self) -> Simulation:
        system = self.add_backbone_posres(self.load_amber_files(), 
                                          self.inpcrd.positions, 
                                          self.prmtop.topology.atoms(), 
                                          self.indices,
                                          self.k)
    
        simulation, integrator = self.setup_sim(system, dt=0.002)
        
        simulation.context.setPositions(self.inpcrd.positions)
        state = simulation.context.getState(getEnergy=True)
        print(f'Energy before minimization: {state.getPotentialEnergy()}')
        simulation.minimizeEnergy()
        state = simulation.context.getState(getEnergy=True)
        print(f'Energy after minimization: {state.getPotentialEnergy()}')
        
        simulation.reporters.append(StateDataReporter(self.eq_log, 
                                                      self.eq_freq, 
                                                      step=True,
                                                      potentialEnergy=True,
                                                      speed=True,
                                                      temperature=True))
        simulation.reporters.append(DCDReporter(f'{self.path}/eq.dcd', self.eq_freq))

        simulation, integrator = self._heating(simulation, integrator)
        simulation = self._equilibrate(simulation)
        
        return simulation

class CustomForcesSimulator(Simulator):
    def __init__(self,
                 path: str,
                 custom_force_objects: list,
                 equil_steps: int=1_250_000, 
                 prod_steps: int=250_000_000, 
                 n_equil_cycles: int=3,
                 reporter_frequency: int=1_000,
                 platform: str='CUDA',
                 device_ids: list[int]=[0],
                 equilibration_force_constant: float=10.):
        super().__init__(path, equil_steps, prod_steps, n_equil_cycles,
                         reporter_frequency, platform, device_ids, 
                         equilibration_force_constant)
        self.custom_forces = custom_force_objects

    def load_amber_files(self) -> System:
        if isinstance(self.inpcrd, str):
            self.inpcrd = AmberInpcrdFile(self.inpcrd)
            try: # This is how it is done in OpenMM 8.0 and on
                self.prmtop = AmberPrmtopFile(self.prmtop, periodicBoxVectors=self.inpcrd.boxVectors)
            except TypeError: # This means we are in OpenMM 7.7 or earlier
                self.prmtop = AmberPrmtopFile(self.prmtop)

        system = self.prmtop.createSystem(nonbondedMethod=PME,
                                          removeCMMotion=False,
                                          nonbondedCutoff=1. * nanometer,
                                          constraints=HBonds,
                                          hydrogenMass=1.5 * amu)

        system = self.add_forces(system)

        return system

    def add_forces(self, system: System) -> System:
        for custom_force in self.custom_forces:
            system.addForce(custom_force)

        return system

class Minimizer:
    def __init__(self,
                 topology: PathLike,
                 out: PathLike='min.pdb',
                 platform: str='OpenCL',
                 device_ids: list[int]=[0]):
        self.topology = Path(topology) if isinstance(topology, str) else topology
        self.path = self.topology.parent
        self.out = self.path / out
        self.platform = platform
        self.properties = {'DeviceIndex': ','.join([str(x) for x in device_ids]),
                           'Precision': 'mixed'}

    def minimize(self) -> None:
        system = self.load_files()
        integrator = LangevinMiddleIntegrator(300*kelvin, 
                                              1/picosecond, 
                                              0.001*picoseconds)
        simulation = Simulation(self.topology, 
                                system, 
                                integrator,
                                self.platform,
                                self.properties)

        simulation.context.setPositions(self.coordinates.positions)

        simulation.minimizeEnergy()

        state = simulation.context.getState(getPositions=True)
        positions = state.getPositions()
        
        PDBFile.writeFile(simulation.topology, 
                          positions, 
                          file=str(self.out), 
                          keepIds=True)

    def load_files(self) -> None:
        if self.topology.suffix in ['.prmtop', '.parm7']:
            system = self.load_amber()
        elif self.topology.suffix == '.top':
            system = self.load_gromacs()
        elif self.topology.suffix == '.pdb':
            system = self.load_pdb()
        else:
            raise FileNotFoundError('No viable simulation input files found'
                                    f'at path: {self.path}!')

        return system
        
    def load_amber(self) -> System:
        inpcrd = list(self.path.glob('*.inpcrd')) + list(self.path.glob('*.rst7'))
        self.coordinates = AmberInpcrdFile(str(inpcrd[0]))
        self.topology = AmberPrmtopFile(str(self.topology))

        system = self.topology.createSystem(nonbondedMethod=NoCutoff,
                                            constraints=HBonds)

        return system

    def load_gromacs(self) -> System:
        gro = list(self.path.glob('*.gro'))[0]
        self.coordinates = GromacsGroFile(str(gro))
        self.topology = GromacsTopFile(str(self.topology), 
                                       includeDir='/usr/local/gromacs/share/gromacs/top')

        system = self.topology.createSystem(nonbondedMethod=NoCutoff, 
                                            constraints=HBonds)

        return system

    def load_pdb(self) -> System:
        self.coordinates = PDBFile(str(self.topology))
        self.topology = self.coordinates.topology
        forcefield = ForceField('amber14-all.xml')

        system = forcefield.createSystem(self.topology, 
                                         nonbondedMethod=NoCutoff,
                                         constraints=HBonds)

        return system



#!/usr/bin/env python
from .mmpbsa import MMPBSA
from .omm_simulator import Simulator, ImplicitSimulator, Minimizer
try:
    from .parsl_settings import (LocalSettings, 
                                 WorkstationSettings, 
                                 PolarisSettings, 
                                 AuroraSettings)
except ImportError:
    pass



from dataclasses import dataclass
import json
import numpy as np
import os
from pathlib import Path
import polars as pl
import subprocess
from typing import Union

PathLike = Union[Path, str]

@dataclass
class MMPBSA_settings:
    top: PathLike
    dcd: PathLike
    selections: list[str]
    first_frame: int = 0
    last_frame: int = -1
    stride: int = 1
    n_cpus: int = 1
    out: str = 'mmpbsa'
    solvent_probe: float = 1.4
    offset: int = 0
    gb_surften: float=0.0072
    gb_surfoff: float=0.

class MMPBSA(MMPBSA_settings):
    """
    This is an experiment in patience. What follows is a reconstruction of the various
    pieces of code that run MM-P(G)BSA from AMBER but written in a more digestible manner
    with actual documentation. Herein we have un-CLI'd what should never have been a
    CLI and piped together the correct pieces of the ambertools ecosystem to perform
    MM-P(G)BSA and that alone. Your trajectory is required to be concatenated into a single
    continuous trajectory - or you can run this serially over each by instancing this class
    for each trajectory you have. In this way we have also disentangled the requirement to
    parallelize by use of MPI, allowing the user to choose their own parallelization/scaling
    scheme.

    Arguments:
        top (PathLike): Input topology for a solvated system. Should match the input trajectory.
        dcd (PathLike): Input trajectory. Can be DCD format or MDCRD already.
        selections (list[str]): A list of residue ID selections for the receptor and ligand
            in that order. Should be formatted for cpptraj (e.g. `:1-10`).
        first_frame (int): Defaults to 0. The first frame of the input trajectory to begin
            the calculations on.
        last_frame (int): Defaults to -1. Optional final frame to cut trajectory at. If -1,
            acts as a flag to run the whole trajectory.
        stride (int): Defaults to 1. The number of frames to stride the trajectory by.
        n_cpus (int): Number of parallel processes
        out (str): The prefix name or path for output files.
        solvent_probe (float): Defaults to 1.4Å. The probe radius to use for SA calculations.
        offset (int): Defaults to 0Å. I don't know what this does.
        gb_surften (float): Defaults to 0.0072.
        gb_suroff (float): Defaults to 0.0.
    """
    def __init__(self,
                 top: PathLike,
                 dcd: PathLike,
                 selections: list[str],
                 first_frame: int=0,
                 last_frame: int=-1,
                 stride: int=1,
                 n_cpus: int=1,
                 out: str='mmpbsa',
                 solvent_probe: float=1.4,
                 offset: int=0,
                 gb_surften: float=0.0072,
                 gb_surfoff: float=0.,
                 amberhome: PathLike='',
                 **kwargs):
        super().__init__(top, dcd, selections, first_frame, last_frame, 
                         out, solvent_probe, offset, gb_surften, gb_surfoff)
        self.top = Path(self.top).resolve()
        self.traj = Path(self.dcd).resolve()
        self.path = self.top.parent
        if out == 'mmpbsa':
            self.path = self.path / 'mmpbsa'
        else:
            self.path = Path(out).resolve()

        self.path.mkdir(exist_ok=True, parents=True)
        os.chdir(str(self.path)) # critical for parallel runs to not overwrite files

        self.cpptraj = 'cpptraj'
        self.mmpbsa_py_energy = 'mmpbsa_py_energy'
        if amberhome: # we are overriding AMBERHOME or using another env's install
            self.cpptraj = Path(amberhome) / self.cpptraj
            self.mmpbsa_py_energy = Path(amberhome) / self.mmpbsa_py_energy

        self.fh = FileHandler(self.top, self.traj, self.path, self.selections, 
                              self.first_frame, self.last_frame, self.stride,
                              self.cpptraj)
        self.analyzer = OutputAnalyzer(self.path, self.gb_surften, self.gb_surfoff)

        for key, value in kwargs.items():
            setattr(self, key, value)

    def run(self) -> None:
        """
        Main logic of MM-PBSA. Computes the SASA with molsurf in cpptraj, and
        the various energy terms for GB/PB using mmpbsa_py_energy from ambertools.
        Finally, parse outputs and collate into a neat form consisting of a json
        of raw data and a plain text file of the binding free energies.
        """
        gb_mdin, pb_mdin = self.write_mdins()

        for (prefix, top, traj, pdb) in self.fh.files:
            self.calculate_sasa(prefix, top, traj)
            self.calculate_energy(prefix, top, traj, pdb, gb_mdin, 'gb')
            self.calculate_energy(prefix, top, traj, pdb, pb_mdin, 'pb')

        self.analyzer.parse_outputs()

    def calculate_sasa(self,
                       pre: str,
                       prm: PathLike,
                       trj: PathLike) -> None:
        """
        Runs the molsurf command in cpptraj to compute the SASA of a given system.
        """
        sasa = self.fh.path / 'sasa.in'
        sasa_in = [
            f'parm {prm}',
            f'trajin {trj}',
            f'molsurf :* out {pre}_surf.dat probe {self.solvent_probe} offset {self.offset}',
            'run',
            'quit'
        ]
        
        self.fh.write_file(sasa_in, sasa)
        subprocess.call(f'{self.cpptraj} -i {sasa}', shell=True)
        sasa.unlink()
    
    def calculate_energy(self,
                         pre: str,
                         prm: PathLike,
                         trj: PathLike,
                         pdb: PathLike, 
                         mdin: PathLike,
                         suf: str) -> None:
        """
        Runs mmpbsa_py_energy, an undocumented binary file which somehow computes the
        energy of a system.
        """
        subprocess.call(f'{self.mmpbsa_py_energy} -O -i {mdin} -p {prm} -c {pdb} -y {trj} -o {pre}_{suf}.mdout', shell=True)
    
    def write_mdins(self) -> tuple[Path, Path]:
        """
        Writes out the configuration files that are to be fed to mmpbsa_py_energy.
        These are also undocumented and I took the parameters from the location
        in which they are hardcoded in ambertools.
        """
        gb = self.fh.path / 'gb_mdin'
        gb_mdin = [
            'GB',
            'igb = 2',
            'extdiel = 78.3',
            'saltcon = 0.10',
            f'surften = {self.gb_surften}',
            'rgbmax = 25.0'
        ]

        self.fh.write_file(gb_mdin, gb)

        pb = self.fh.path / 'pb_mdin'
        pb_mdin = [
            'PB',
            'inp = 2',
            'smoothopt = 1',
            'radiopt = 0',
            'npbopt = 0',
            'solvopt = 1',
            'maxitn = 1000',
            'nfocus = 2',
            'bcopt = 5',
            'eneopt = 2',
            'fscale = 8',
            'epsin = 1.0',
            'epsout = 80.0',
            'istrng = 0.10',
            'dprob = 1.4',
            'iprob = 2.0',
            'accept = 0.001',
            'fillratio = 4.0',
            'space = 0.5',
            'cutnb = 0',
            'sprob = 0.557',
            'cavity_surften = 0.0378',
            'cavity_offset = -0.5692'
        ]

        self.fh.write_file(pb_mdin, pb)

        return gb, pb


class OutputAnalyzer:
    """
    Analyzes the outputs from an MM-PBSA run. Stores data in a Polars dataframe
    internally, and writes out data in the form of json/plain text.
    """
    def __init__(self, 
                 path: PathLike,
                 surface_tension: float=0.0072,
                 sasa_offset: float=0.,
                 _tolerance: float = 0.005):
        self.path = Path(path)
        self.surften = surface_tension
        self.offset = sasa_offset
        self.tolerance = _tolerance

        self.systems = ['receptor', 'ligand', 'complex']
        self.levels = ['gb', 'pb']

        self.solvent_contributions = ['EGB', 'ESURF', 'EPB', 'ECAVITY']

    def parse_outputs(self) -> None:
        """
        Parse all the output files.
        """        
        self.gb = pl.DataFrame()
        self.pb = pl.DataFrame()
        for system in self.systems:
            E_sasa = self.read_sasa(self.path / f'{system}_surf.dat')
            E_gb = self.read_GB(self.path / f'{system}_gb.mdout', system)
            E_pb = self.read_PB(self.path / f'{system}_pb.mdout', system)

            E_gb = E_gb.drop('ESURF').with_columns(E_sasa)

            self.gb = pl.concat([self.gb, E_gb], how='vertical')
            self.pb = pl.concat([self.pb, E_pb], how='vertical')
        
        all_cols = list(set(self.gb.columns + self.pb.columns))
        self.contributions = {
                'G gas': [col for col in all_cols
                          if col not in self.solvent_contributions], 
                'G solv': [col for col in all_cols
                          if col in self.solvent_contributions]
            }
        
        self.check_bonded_terms()
        self.generate_summary()
        self.compute_dG()

    def read_sasa(self,
                  _file: PathLike) -> np.ndarray:
        """
        Reads in the results of the cpptraj SASA calculation and returns the
        per-frame SASA scaled by a hardcoded value for surface tension that is
        a mostly undocumented heuristic.

        Arguments:
            _file (PathLike): Path to a file containing the SASA data.
        Returns:
            (np.ndarray): A numpy array of the per-frame rescaled SASA energies.
        """
        sasa = []
        for line in open(_file).readlines()[1:]:
            sasa.append(line.split()[-1].strip())

        return pl.Series('ESURF', np.array(sasa, dtype=float) * self.surften + self.offset)

    def read_GB(self,
                _file: PathLike,
                system: str) -> pl.DataFrame:
        """
        Read in the GB mdout files and returns a Polars dataframe of the values
        for each term for every frame. Also adds a `system` label to more easily
        compute summary statistics later.
        """
        gb_terms = ['BOND', 'ANGLE', 'DIHED', 'VDWAALS', 'EEL',
                    'EGB', '1-4 VDW', '1-4 EEL', 'RESTRAINT', 'ESURF']
        data = {gb_term: [] for gb_term in gb_terms}
        
        lines = open(_file, 'r').readlines()

        return self.parse_energy_file(lines, data, system)

    def read_PB(self,
                _file: PathLike,
                system: str) -> pl.DataFrame:
        """
        Read in the PB mdout files and returns a Polars dataframe of the values
        for each term for every frame. Also adds a `system` label to more easily
        compute summary statistics later.
        """
        pb_terms = ['BOND', 'ANGLE', 'DIHED', 'VDWAALS', 'EEL',
                    'EPB', '1-4 VDW', '1-4 EEL', 'RESTRAINT',
                    'ECAVITY', 'EDISPER']
        data = {pb_term: [] for pb_term in pb_terms}

        lines = open(_file, 'r').readlines()

        return self.parse_energy_file(lines, data, system)
    
    def parse_energy_file(self,
                          file_contents: list[str],
                          data: dict[str, list],
                          system: str) -> pl.DataFrame:
        """
        Parses the contents of an energy calculation using a dictionary of
        energy terms to extract theory-level observables (e.g. EGB vs EPB).

        Arguments:
            file_contents (list[str]): A list of each line from an energy calculation.
            data (dict[str, list]): The relevant energy terms to be scraped from input.
            system (str): The name of the system which will be included as an additional
                kv pair in the returned dataframe. This ensures we can track which portion
                of the calculation we are accounting for (e.g. complex, receptor, ligand).
        Returns:
            (pl.DataFrame): A Polars dataframe of shape (n_frames, n_calculations + system).
        """
        idx = 0
        n_frames = 0
        while idx < len(file_contents):
            if file_contents[idx].startswith(' BOND'):
                for _ in range(4): # number of lines to read. DO NOT CHANGE!!!
                    line = file_contents[idx]
                    parsed = self.parse_line(line)
                    for key, val in parsed:
                        data[key].append(val)

                    idx += 1

            if 'Processing frame' in file_contents[idx]:
                n_frames = int(file_contents[idx].strip().split()[-1])

            idx +=1 

        data['system'] = [system] * n_frames
        
        return pl.DataFrame(
            {key: np.array(val) for key, val in data.items()}
        )

    def check_bonded_terms(self) -> None:
        bonded = ['BOND', 'ANGLE', 'DIHED', '1-4 VDW', '1-4 EEL']
        
        for theory_level in (self.gb, self.pb):
            a = theory_level.filter(pl.col('system') == 'receptor')
            b = theory_level.filter(pl.col('system') == 'ligand')
            c = theory_level.filter(pl.col('system') == 'complex')

            a = a.select(pl.col([col for col in a.columns if col in bonded])).to_numpy()
            b = b.select(pl.col([col for col in b.columns if col in bonded])).to_numpy()
            c = c.select(pl.col([col for col in c.columns if col in bonded])).to_numpy()

            diffs = np.array(c - b - a)
            if np.where(diffs >= self.tolerance)[0].size > 0:
                raise ValueError('Bonded terms for receptor + ligand != complex!')

        remove = ['RESTRAINT', 'EDISPER']
        self.gb = self.gb.select(
            pl.col([col for col in self.gb.columns if col not in remove])
        )
        self.pb = self.pb.select(
            pl.col([col for col in self.pb.columns if col not in remove])
        )

        self.n_frames = self.gb.height
        self.square_root_N = np.sqrt(self.n_frames)

    def generate_summary(self) -> None:
        full_statistics = {sys: {} for sys in self.systems}
        for theory, level in zip([self.gb, self.pb], self.levels):
            for system in self.systems:
                sys = theory.filter(pl.col('system') == system).drop('system')

                stats = {}
                for col in sys.columns:
                    mean = sys.select(pl.mean(col)).item()
                    stdev = sys.select(pl.std(col)).item()
                    
                    stats[col] = {'mean': mean, 
                                  'std': stdev, 
                                  'err': stdev / self.square_root_N}

                for energy, contributors in self.contributions.items():
                    pooled_data = sys.select(
                        pl.col([col for col in sys.columns if col in contributors])
                    ).to_numpy().flatten()

                    stats[energy] = {'mean': np.mean(pooled_data),
                                     'std': np.std(pooled_data),
                                     'err': np.std(pooled_data) / self.square_root_N}

                total_data = sys.to_numpy().flatten()
                stats['total'] = {'mean': np.mean(total_data),
                                  'std': np.std(total_data),
                                  'err': np.std(total_data) / self.square_root_N}

                full_statistics[system][level] = stats
        
        with open('statistics.json', 'w') as fout:
            json.dump(full_statistics, fout, indent=4)

    def compute_dG(self) -> None:
        differences = []
        for theory, level in zip([self.gb, self.pb], self.levels):
            diff_cols = [col for col in theory.columns if col != 'system']
            diff_arr = theory.filter(pl.col('system') == 'complex').drop('system').to_numpy()
            for system in self.systems[:2]:
                diff_arr -= theory.filter(pl.col('system') == system).drop('system').to_numpy()

            means = np.mean(diff_arr, axis=0)
            stds = np.std(diff_arr, axis=0)
            errs = stds / self.square_root_N

            gas_solv_phase = []
            for energy, contributors in self.contributions.items():
                indices = [i for i, diff_col in enumerate(diff_cols) 
                           if diff_col in contributors]
                contribution = np.sum(diff_arr[:, indices], axis=1)
                gas_solv_phase.append(contribution)

                diff_cols.append(energy)
                means = np.concatenate((means, [np.mean(contribution)]))
                stds = np.concatenate((stds, [np.std(contribution)]))
                errs = np.concatenate((errs, [np.std(contribution) / self.square_root_N]))
            
            diff_cols.append('∆G Binding')
            total = np.sum(np.vstack(gas_solv_phase), axis=0)
            
            means = np.concatenate((means, [np.mean(total)]))
            stds = np.concatenate((stds, [np.std(total)]))
            errs = np.concatenate((errs, [np.std(total) / self.square_root_N]))

            data = np.vstack((means, stds, errs))
            
            differences.append(pl.DataFrame(
                {diff_cols[i]: data[:,i] for i in range(len(diff_cols))}
            ))

        self.pretty_print(differences)

    def pretty_print(self,
                     dfs: list[pl.DataFrame]) -> None:
        print_statement = []
        for df, level in zip(dfs, ['Generalized Born ', 'Poisson Boltzmann']):
            print_statement += [
                f'{" ":<20}=========================',
                f'{" ":<20}=== {level} ===',
                f'{" ":<20}=========================',
                'Energy Component    Average         Std. Dev.       Std. Err. of Mean',
                '---------------------------------------------------------------------'
            ]
            for col in df.columns:
                mean, std, err = [x.item() for x in df.select(pl.col(col)).to_numpy()]
                if abs(mean) <= self.tolerance:
                    continue

                if col in ['G gas', '∆G Binding']:
                    print_statement.append('')

                print_statement.append(f'{col:<20}{mean:<16.3f}{std:<16.3f}{err:<16.3f}')

        print_statement = '\n'.join(print_statement)
        with open('deltaG.txt', 'w') as fout:
            fout.write(print_statement)
        
        print(print_statement)

    @staticmethod
    def parse_line(line) -> tuple[list[str], list[float]]:
        eq_split = line.split('=')
        
        if len(eq_split) == 2:
            splits = [eq_spl.strip() for eq_spl in eq_split]
        else:
            splits = [eq_split[0].strip()]

            for i in range(1, len(eq_split) - 1):
                splits += [spl.strip() for spl in eq_split[i].strip().split('  ')]

            splits += [eq_split[-1].strip()]
        
        keys = splits[::2]
        vals = np.array(splits[1::2], dtype=float)
        
        return zip(keys, vals)


class FileHandler:
    """
    Performs preprocessing for MM-PBSA runs and manages the pathing to all file
    inputs. Additionally used to write out various cpptraj input files by the
    MMPBSA class.
    """
    def __init__(self,
                 top: Path,
                 traj: Path,
                 path: Path,
                 sels: list[str],
                 first: int,
                 last: int,
                 stride: int,
                 cpptraj_binary: PathLike):
        self.top = top
        self.traj = traj
        self.path = path
        self.selections = sels
        self.ff = first
        self.lf = last
        self.stride = stride
        self.cpptraj = cpptraj_binary

        self.prepare_topologies()
        self.prepare_trajectories()

    def prepare_topologies(self) -> None:
        """
        Slices out each sub-topology for the desolvated complex, receptor and
        ligand using cpptraj due to the difficulty of working with AMBER FF
        files otherwise (including PARMED).
        """
        self.topologies = [
            self.path / 'complex.prmtop',
            self.path / 'receptor.prmtop',
            self.path / 'ligand.prmtop'
        ]

        print(self.topologies)
        
        cpptraj_in = [
            f'parm {self.top}',
            'parmstrip :Na+,Cl-,WAT',
            'parmbox nobox',
            f'parmwrite out {self.topologies[0]}',
            'run',
            'clear all',
            f'parm {self.topologies[0]}',
            f'parmstrip {self.selections[0]}',
            f'parmwrite out {self.topologies[1]}',
            'run',
            'clear all',
            f'parm {self.topologies[0]}',
            f'parmstrip {self.selections[1]}',
            f'parmwrite out {self.topologies[2]}',
            'run',
            'quit'
        ]
        
        script = self.path  / 'cpptraj.in'
        self.write_file('\n'.join(cpptraj_in), script)
        subprocess.call(f'{self.cpptraj} -i {script}', shell=True)
        script.unlink()
        
    def prepare_trajectories(self) -> None:
        """
        Converts DCD trajectory to AMBER CRD format which is explicitly
        required by MM-G(P)BSA.
        """
        self.trajectories = [path.with_suffix('.crd') for path in self.topologies]
        self.pdbs = [path.with_suffix('.pdb') for path in self.topologies]
        
        frame_control = f'start {self.ff}'

        if self.lf > -1:
            frame_control += f'stop {self.lf}'
        
        frame_control += f'offset {self.stride}'
        
        cpptraj_in = [
            f'parm {self.top}', 
            f'trajin {self.traj}',
            f'trajout {self.traj.with_suffix(".crd")} crd {frame_control}',
            'run',
            'clear all',
        ]

        self.traj = self.traj.with_suffix('.crd')

        cpptraj_in += [
            f'parm {self.top}', 
            f'trajin {self.traj}',
            'strip :WAT,Na+,Cl*',
            'autoimage',
            f'rmsd !(:WAT,Cl*,CIO,Cs+,IB,K*,Li+,MG*,Na+,Rb+,CS,RB,NA,F,CL) mass first',
            f'trajout {self.trajectories[0]} crd nobox',
            f'trajout {self.pdbs[0]} pdb onlyframes 1',
            'run',
            'clear all',
            f'parm {self.topologies[0]}', 
            f'trajin {self.trajectories[0]}',
            f'strip {self.selections[0]}',
            f'trajout {self.trajectories[1]} crd',
            f'trajout {self.pdbs[1]} pdb onlyframes 1',
            'run',
            'clear all',
            f'parm {self.topologies[0]}', 
            f'trajin {self.trajectories[0]}',
            f'strip {self.selections[1]}',
            f'trajout {self.trajectories[2]} crd',
            f'trajout {self.pdbs[2]} pdb onlyframes 1',
            'run',
            'quit'
        ]

        name = self.path / 'mdcrd.in'
        self.write_file('\n'.join(cpptraj_in), name)
        subprocess.call(f'{self.cpptraj} -i {name}', shell=True)

        name.unlink()

    @property
    def files(self) -> tuple[list[str]]:
        """
        Returns a zip generator containing the output paths, topologies,
        trajectories and pdbs for each system. This is done to ensure we
        have the correct order for housekeeping reasons.
        """
        _order = [self.path / prefix for prefix in ['complex', 'receptor', 'ligand']]
        return zip(_order, self.topologies, self.trajectories, self.pdbs)

    @staticmethod
    def write_file(lines: list[str],
                   filepath: PathLike) -> None:
        """
        Given an input of either a list of strings or a single string,
        write input to file. If a list, join by newline characters.

        Arguments:
            lines (list[str]): Input to be written to file.
            filepath (PathLike): Path to the file to be written.
        """
        if isinstance(lines, list):
            lines = '\n'.join(lines)
        with open(str(filepath), 'w') as f:
            f.write(lines)



from abc import ABC, abstractmethod
from copy import deepcopy
from openmm import *
from openmm.app import *
from openmm.unit import *
import MDAnalysis as mda
from MDAnalysis.analysis.distances import contact_matrix
import mdtraj as md
import numpy as np
import parmed as pmd
from pathlib import Path
from pdbfixer import PDBFixer
import pickle
import gc
from tqdm import tqdm
from typing import Dict, List, Tuple, Union

PathLike = Union[Path, str]

class InteractionEnergy(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def compute(self):
        pass

    @abstractmethod
    def energy(self):
        pass

    @abstractmethod
    def get_selection(self):
        pass

class StaticInteractionEnergy(InteractionEnergy):
    """
    Computes the linear interaction energy between specified chain and other simulation
    components. Can specify a range of residues in chain to limit calculation to. Works on
    a static model but can be adapted to run on dynamics data.

    Inputs:
        pdb (str): Path to input PDB file
        chain (str): Defaults to A. The chain for which to compute the energy between.
            Computes energy between this chain and all other components in PDB file.
        first_residue (int, None): If set, will restrict the calculation to residues
            beginning with resid `first_residue`.
        last_residue (int, None): If set, will restrict the calculation to residues
            ending with resid `last_residue`.
    """
    def __init__(self, pdb: str, chain: str='A', platform: str='CUDA',
                 first_residue: Union[int, None]=None, 
                 last_residue: Union[int, None]=None):
        self.pdb = pdb
        self.chain = chain
        self.platform = Platform.getPlatformByName(platform)
        self.first = first_residue
        self.last = last_residue
        
    def get_system(self) -> None:
        pdb = PDBFile(self.pdb)
        positions, topology = pdb.positions, pdb.topology
        forcefield = ForceField('amber14-all.xml', 'implicit/gbn2.xml')
        try:
            system = forcefield.createSystem(topology,
                                             soluteDielectric=1.,
                                             solventDielectric=80.)
        except ValueError:
            positions, topology = self.fix_pdb()
            system = forcefield.createSystem(topology,
                                             soluteDielectric=1.,
                                             solventDielectric=80.)

        self.positions = positions
        self.get_selection(topology)

        return system

    def compute(self, positions: Union[np.ndarray, None]=None) -> None:
        self.lj = None
        self.coulomb = None

        system = self.get_system()
        if positions is None:
            positions = self.positions
            
        for force in system.getForces():
            if isinstance(force, NonbondedForce):
                force.setForceGroup(0)
                force.addGlobalParameter("solute_coulomb_scale", 1)
                force.addGlobalParameter("solute_lj_scale", 1)
                force.addGlobalParameter("solvent_coulomb_scale", 1)
                force.addGlobalParameter("solvent_lj_scale", 1)

                for i in range(force.getNumParticles()):
                    charge, sigma, epsilon = force.getParticleParameters(i)
                    force.setParticleParameters(i, 0, 0, 0)
                    if i in self.selection:
                        force.addParticleParameterOffset("solute_coulomb_scale", i, charge, 0, 0)
                        force.addParticleParameterOffset("solute_lj_scale", i, 0, sigma, epsilon)
                    else:
                        force.addParticleParameterOffset("solvent_coulomb_scale", i, charge, 0, 0)
                        force.addParticleParameterOffset("solvent_lj_scale", i, 0, sigma, epsilon)

                for i in range(force.getNumExceptions()):
                    p1, p2, chargeProd, sigma, epsilon = force.getExceptionParameters(i)
                    force.setExceptionParameters(i, p1, p2, 0, 0, 0)

            else:
                force.setForceGroup(2)
        
        integrator = VerletIntegrator(0.001*picosecond)

        context = Context(system, integrator, self.platform)
        context.setPositions(positions)
        
        total_coulomb = self.energy(context, 1, 0, 1, 0)
        solute_coulomb = self.energy(context, 1, 0, 0, 0)
        solvent_coulomb = self.energy(context, 0, 0, 1, 0)
        total_lj = self.energy(context, 0, 1, 0, 1)
        solute_lj = self.energy(context, 0, 1, 0, 0)
        solvent_lj = self.energy(context, 0, 0, 0, 1)
        
        coul_final = total_coulomb - solute_coulomb - solvent_coulomb
        lj_final = total_lj - solute_lj - solvent_lj

        self.coulomb = coul_final.value_in_unit(kilocalories_per_mole)
        self.lj = lj_final.value_in_unit(kilocalories_per_mole)
    
    def get_selection(self, topology) -> None:
        if self.first is None and self.last is None:
            selection = [a.index 
                        for a in topology.atoms() 
                        if a.residue.chain.id == self.chain]
        elif self.first is not None and self.last is None:
            selection = [a.index
                        for a in topology.atoms()
                        if a.residue.chain.id == self.chain 
                        and int(self.first) <= int(a.residue.id)]
        elif self.first is None:
            selection = [a.index
                        for a in topology.atoms()
                        if a.residue.chain.id == self.chain 
                        and int(self.last) >= int(a.residue.id)]
        else:
            selection = [a.index
                        for a in topology.atoms()
                        if a.residue.chain.id == self.chain 
                        and int(self.first) <= int(a.residue.id) <= int(self.last)]

        self.selection = selection

    def fix_pdb(self):
        fixer = PDBFixer(filename=self.pdb)
        fixer.findMissingResidues()
        fixer.findMissingAtoms()
        fixer.addMissingAtoms()
        fixer.addMissingHydrogens(7.0)

        return fixer.positions, fixer.topology
    
    @property
    def interactions(self) -> np.ndarray:
        return np.vstack([self.lj, self.coulomb])

    @staticmethod
    def energy(context, solute_coulomb_scale: int=0, solute_lj_scale: int=0, 
               solvent_coulomb_scale: int=0, 
               solvent_lj_scale: int=0) -> float:
        context.setParameter("solute_coulomb_scale", solute_coulomb_scale)
        context.setParameter("solute_lj_scale", solute_lj_scale)
        context.setParameter("solvent_coulomb_scale", solvent_coulomb_scale)
        context.setParameter("solvent_lj_scale", solvent_lj_scale)
        return context.getState(getEnergy=True, groups={0}).getPotentialEnergy()

class InteractionEnergyFrame(StaticInteractionEnergy):
    def __init__(self, system: System, top: Topology, 
                 chain: str='A', platform: str='CUDA',
                 first_residue: Union[int, None]=None, 
                 last_residue: Union[int, None]=None):
        super().__init__('', chain, platform, first_residue, last_residue)
        self.system = system
        self.top = top

    def get_system(self):
        self.get_selection(self.top)
        return self.system


class DynamicInteractionEnergy:
    def __init__(self, top: str, traj: str, stride: int=1, 
                 chain: str='A', platform: str='CUDA',
                 first_residue: Union[int, None]=None,
                 last_residue: Union[int, None]=None,
                 progress_bar: bool=False):
        self.system = self.build_system(top, traj)
        self.coordinates = self.load_traj(top, traj)
        self.stride = stride
        self.progress = progress_bar

        self.IE = InteractionEnergyFrame(self.system, self.top, chain, 
                                         platform, first_residue, last_residue)

    def compute_energies(self) -> None:
        n_frames = self.coordinates.shape[0] // self.stride
        self.energies = np.zeros((n_frames, 2))
        
        if self.progress:
            pbar = tqdm(total=n_frames, position=0, leave=False)

        for i in range(n_frames):
            fr = i * self.stride
            self.IE.compute(self.coordinates[fr, :, :])
            self.energies[i, 0] = self.IE.lj
            self.energies[i, 1] = self.IE.coulomb

            if self.progress:
                pbar.update(1)

        if self.progress:
            pbar.close()
    
    def build_system(self, top: str, traj: str) -> System:
        if top[-3:] == 'pdb':
            top = PDBFile(top).topology
            self.top = top
            forcefield = ForceField('amber14-all.xml', 'implicit/gbn2.xml')
            return forcefield.createSystem(top, 
                                           soluteDielectric=1., 
                                           solventDielectric=80.)
        elif top[-6:] == 'prmtop':
            top = AmberPrmtopFile(top)
            self.top = top
            return top.createSystem(nonbondedMethod=CutoffNonPeriodic,
                                    nonbondedCutoff=2. * nanometers,
                                    constraints=HBonds)
        else:
            raise NotImplementedError(f'Error! Topology type {top} not implemented!')

    def load_traj(self, top: str, traj: str) -> np.ndarray:
        return md.load(traj, top=top).xyz

    def setup_pbar(self) -> None:
        self.pbar = tqdm(total=self.coordinates.shape[0], position=0, leave=False)



import json
import numpy as np
from pathlib import Path
import polars as pl
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.metrics import silhouette_score
from tqdm import tqdm
from typing import Any, Type, TypeVar, Union

PathLike = Union[Path, str]
_T = TypeVar('_T')

class GenericDataloader:
    """
    Loads any generic data stored in numpy arrays and stores the full
    dataset. Capable of loading data with variable row lengths but must
    be consistent in the columnar dimension.
    """
    def __init__(self,
                 data_files: list[PathLike]):
        self.files = data_files
        self.load_data()

    def load_data(self):
        """
        Loads data into one large array.
        """
        self.data_array = []
        self.shapes = []
        for f in self.files:
            temp = np.load(str(f))
            self.shapes.append(temp.shape)
            self.data_array.append(temp)
        
        self.data_array = np.vstack(self.data_array)
        if len(self.data_array) > 2:
            x, *y = self.data_array.shape
            shape2 = 1
            for shape in y:
                shape2 *= shape

            self.data_array = self.data_array.reshape(x, shape2)

    @property
    def data(self):
        return self.data_array

    @property
    def shape(self):
        if len(set(self.shapes)) == 1:
            return self.shapes[0]
        
        return self.shapes


class PeriodicDataloader(GenericDataloader):
    """
    Decomposes periodic data using sin and cos, returning double the features.
    """
    def __init__(self,
                 data_files: list[PathLike]):
        super().__init__(data_files)
        
    def load_data(self):
        """
        Loads file of input periodic data.
        """
        self.data_array = []
        self.shapes = []
        for f in self.files:
            temp = self.remove_periodicity(np.load(str(f)))
            
            self.shapes.append(temp.shape)
            self.data_array.append(temp)

    def remove_periodicity(self,
                           arr: np.ndarray) -> np.ndarray:
        """
        Removes periodicity from each feature using sin and cos.
        """
        n_features = arr.shape[1]
        return_arr = np.zeros((arr.shape[0], n_features * 2))
        
        for i in range(n_features):
            return_arr[:, 2*i]   = np.cos(arr[:, i])
            return_arr[:, 2*i+1] = np.sin(arr[:, i])

        return return_arr


class AutoKMeans:
    """
    Performs automatic clustering using KMeans++ including dimensionality reduction
    of the feature space.
    """
    def __init__(self,
                 data_directory: PathLike,
                 pattern: str='',
                 dataloader: Type[_T]=GenericDataloader,
                 max_clusters: int=10,
                 stride: int=1,
                 reduction_algorithm: str='PCA',
                 reduction_kws: dict[str, Any]={'n_components': 2}):
        self.data_dir = Path(data_directory) if isinstance(data_directory, str) else data_directory
        self.dataloader = dataloader(list(self.data_dir.glob(f'{pattern}*.npy')))
        self.data = self.dataloader.data
        self.shape = self.dataloader.shape
        
        self.n_clusters = max_clusters
        self.stride = stride

        self.decomposition = Decomposition(reduction_algorithm, **reduction_kws)
    
    def run(self) -> None:
        """
        Runs the automated clustering workflow.
        """
        self.reduce_dimensionality()
        self.sweep_n_clusters([n for n in range(2, self.n_clusters, self.stride)])
        self.map_centers_to_frames()
        self.save_centers()
        self.save_labels()

    def reduce_dimensionality(self) -> None:
        """
        Performs dimensionality reduction using decomposer of choice.
        """
        self.reduced = self.decomposition.fit_transform(self.data)

    def sweep_n_clusters(self,
                         n_clusters: list[int]) -> None:
        """
        Uses silhouette score to perform a parameter sweep over number of clusters.
        Stores the cluster centers for the best performing parameterization.
        """
        best_centers = None
        best_labels = None
        best_score = 0.
        for n in tqdm(n_clusters, total=len(n_clusters), position=0, 
                      leave=False, desc='Sweeping `n_clusters`'):
            clusterer = KMeans(n_clusters=n)
            labels = clusterer.fit_predict(self.reduced)
            average_score = silhouette_score(self.reduced, labels)

            if average_score > best_score:
                best_centers = clusterer.cluster_centers_
                best_labels = labels
                best_score = average_score

        self.centers = best_centers
        self.labels = best_labels

    def map_centers_to_frames(self) -> None:
        """
        Finds and stores the data point which lies closest to the cluster center 
        for each cluster.
        """
        cluster_centers = {i: None for i in range(len(self.centers))}
        for i, center in enumerate(self.centers):
            closest = 100.
            for p, point in enumerate(self.reduced):
                if (dist := np.linalg.norm(point - center)) < closest:
                    rep = p // self.shape[0]
                    frame = p % self.shape[0]
                    cluster_centers[i] = (rep, frame)
                    closest = dist

        self.cluster_centers = cluster_centers

    def save_centers(self) -> None:
        """
        Saves out cluster centers as a json file.
        """
        with open(str(self.data_dir / 'cluster_centers.json'), 'w') as fout:
            json.dump(self.cluster_centers, fout, indent=4)

    def save_labels(self) -> None:
        """
        Generates a polars dataframe containing system, frame and cluster label 
        assignments and saves to a parquet file.
        """
        files = self.dataloader.files
        if isinstance(self.dataloader.shape, tuple):
            shapes = [self.dataloader.shape[0]] * len(files)
        else:
            shapes = [shape[0] for shape in self.dataloader.shape]
        
        df = pl.DataFrame()
        for file, shape in zip(files, shapes):
            temp = pl.DataFrame({'system': [file.name] * shape, 'frame': np.arange(shape)})
            df = pl.concat([df, temp], how='vertical')

        df = df.with_columns(pl.Series(self.labels).alias('cluster'))

        df.write_parquet(str(self.data_dir / 'cluster_assignments.parquet'))


class Decomposition:
    """
    Thin wrapper for various dimensionality reduction algorithms. Uses scikit-learn style
    methods like `fit` and `fit_transform`.
    """
    def __init__(self,
                 algorithm: str,
                 **kwargs):
        algorithms = {
            'PCA': PCA,
            'TICA': None,
            'UMAP': None
        }

        self.decomposer = algorithms[algorithm](**kwargs)
    
    def fit(self,
            X: np.ndarray) -> None:
        self.decomposer.fit(X)

    def transform(self,
                  X: np.ndarray) -> np.ndarray:
        return self.decomposer.transform(X)

    def fit_transform(self,
                      X: np.ndarray) -> np.ndarray:
        return self.decomposer.fit_transform(X)



import MDAnalysis as mda
from MDAnalysis.analysis.base import AnalysisBase
from MDAnalysis.core import groups
from MDAnalysis.guesser.tables import vdwradii
import numpy as np
from scipy.spatial import KDTree
import warnings

warnings.filterwarnings('ignore')

class SASA(AnalysisBase):
    def __init__(self, 
                 ag: mda.AtomGroup,
                 probe_radius: float=1.4,
                 n_points: int=256,
                 **kwargs):
        if isinstance(ag, groups.UpdatingAtomGroup):
            raise TypeError('UpdatingAtomGroups are not valid for SASA!')
        
        super(SASA, self).__init__(ag.universe.trajectory, **kwargs)
        
        if not hasattr(ag, 'elements'):
            raise ValueError('Cannot assign atomic radii:'
                             'Universe has no `elements` property!')

        self.ag = ag
        self.probe_radius = probe_radius
        self.n_points = n_points

        self.radii_dict = dict()
        self.radii_dict.update(vdwradii)

        self.radii = np.vectorize(self.radii_dict.get)(self.ag.elements)
        self.radii += self.probe_radius
        self.max_radii = 2 * np.max(self.radii)

        self.sphere = self.get_sphere()

    def get_sphere(self) -> np.ndarray:
        dl = np.pi * (3 - np.sqrt(5))
        dz = 2. / self.n_points
        longitude = 0
        z = 1 - dz / 2

        xyz = np.zeros((self.n_points, 3), dtype=np.float32)
        for i in range(self.n_points):
            r = np.sqrt(1 - z**2)
            xyz[i, :] = [np.cos(longitude) * r, np.sin(longitude) * r, z]

            z -= dz
            longitude += dl

        return xyz

    def measure_sasa(self,
                     ag: mda.AtomGroup) -> float:
        kdt = KDTree(ag.positions, 10)

        points = np.zeros(ag.n_atoms)
        for i in range(ag.n_atoms):
            sphere = self.sphere.copy() * self.radii[i]
            sphere += ag.positions[i]
            available = self.points_available.copy()
            kdt_sphere = KDTree(sphere, 10)

            for j in kdt.query_ball_point(ag.positions[i], 
                                          self.max_radii, 
                                          workers=-1):
                if j == i:
                    continue
                if self.radii[j] < (self.radii[i] + self.radii[j]):
                    available -= {
                        n for n in kdt_sphere.query_ball_point(
                            self.ag.positions[j],
                            self.radii[j]
                        )
                    }

            points[i] = len(available)

        return 4 * np.pi * self.radii**2 * points / self.n_points

    def _prepare(self):
        self.results.sasa = np.zeros(self.ag.n_residues)

        self.points_available = set(range(self.n_points))

    def _single_frame(self):
        area = self.measure_sasa(self.ag)
        result = np.zeros(self.ag.n_residues)
        for i, atom in enumerate(self.ag.atoms):
            result[atom.resid - 1] += area[i]
        
        self.results.sasa += result

    def _conclude(self):
        if self.n_frames != 0:
            self.results.sasa /= self.n_frames
            

class RelativeSASA(SASA):
    def __init__(self, 
                 ag: mda.AtomGroup,
                 probe_radius: float=1.4,
                 n_points: int=256,
                 **kwargs):
        if not hasattr(ag, 'bonds'):
            raise ValueError('Universe has no `bonds` property!')
        super(RelativeSASA, self).__init__(ag, probe_radius, n_points, **kwargs)

    def _prepare(self):
        self.results.sasa = np.zeros(self.ag.n_residues)
        self.results.relative_area = np.zeros(self.ag.n_residues)

        self.points_available = set(range(self.n_points))

    def _single_frame(self):
        area = self.measure_sasa(self.ag)
        result = np.zeros(self.ag.n_residues)
        for i, atom in enumerate(self.ag.atoms):
            result[atom.resid - 1] += area[i]
        
        self.results.sasa += result

        for res_index in self.ag.residues.resindices:
            tri_peptide = self.ag.select_atoms(f'byres (bonded resindex {res_index})')

            if len(tri_peptide) == 0:
                continue

            tri_pep_area = self.measure_sasa(tri_peptide)
            exposed_area = sum([
                a for a, _id in zip(tri_pep_area, tri_peptide.resindices)
                if _id == res_index
            ])

            if exposed_area != 0.:
                result[resindex] /= exposed_area

        self.results.relative_area += np.array([result[_id] for _id in self.ag.residues.resids])

    def _conclude(self):
        if self.n_frames != 0:
            self.results.sasa /= self.n_frames
            self.results.relative_area /= self.n_frames



import MDAnalysis as mda
import numpy as np
from pathlib import Path
import shutil
from typing import Callable, Union

OptPath = Union[Path, str, None]

class EmbedData:
    """
    Embeds given data into the beta-factor column of PDB. Writes out to same
    path as input PDB and backs up old PDB file, unless an output path is 
    explicitly provided. Embedding data should be provided as a dictionary where
    the keys are MDAnalysis selection strings and the values are numpy arrays
    of shape (n_frames, n_residues, n_datapoints) or (n_residues, n_datapoints).
    """
    def __init__(self,
                 pdb: Path,
                 embedding_dict: dict[str, np.ndarray],
                 out: OptPath=None):
        self.pdb = pdb if isinstance(pdb, Path) else Path(pdb)
        self.embeddings = embedding_dict
        self.out = out if out is not None else self.pdb
        
        self.u = mda.Universe(str(self.pdb))

    def embed(self):
        for sel, data in self.embeddings.items():
            self.embed_selection(sel, data)

        self.write_new_pdb()

    def embed_selection(self,
                        selection: str,
                        data: np.ndarray) -> None:
        sel = self.u.select_atoms(selection)

        for residue, datum in zip(sel.residues, data):
            residue.atoms.tempfactors = np.full(residue.atoms.tempfactors.shape, datum)
    
    def write_new_pdb(self) -> None:
        if self.out.exists():
            if not self.pdb.with_suffix('.orig.pdb').exists():
                shutil.copyfile(str(self.pdb), str(self.pdb.with_suffix('.orig.pdb')))

        with mda.Writer(str(self.out)) as W:
            W.write(self.u.atoms)


class EmbedEnergyData(EmbedData):
    """

    """
    def __init__(self,
                 pdb: Path,
                 embedding_dict: dict[str, np.ndarray],
                 out: OptPath=None):
        super().__init__(pdb, embedding_dict, out)
        self.embeddings = self.preprocess()

    def preprocess(self) -> dict[str, np.ndarray]:
        new_embeddings = dict()
        all_data = []
        for sel, data in self.embeddings.items():
            sanitized = self.sanitize_data(data)
            all_data.append(sanitized)

        rescaling_factor = np.min(np.concatenate(all_data))
        for sel, data in self.embeddings.items():
            sanitized = self.sanitize_data(data)
            rescaled = sanitized / rescaling_factor
            rescaled[np.where(rescaled > 1.)] = 1.
            new_embeddings[sel] = rescaled

        return new_embeddings

    @staticmethod
    def sanitize_data(data: np.ndarray,) -> np.ndarray:
        if len(data.shape) > 2:
            data = np.mean(data, axis=0)

        if data.shape[1] > 1:
            data = np.sum(data, axis=1)
        
        return data



from pathlib import Path
from typing import Union

from .build_amber import ExplicitSolvent, ImplicitSolvent
from .build_interface import InterfaceBuilder

try:
    from .build_ligand import LigandBuilder, PLINDERBuilder, ComplexBuilder
except ImportError: # no rdkit in environment
    pass

PathLike = Union[Path, str]

def convert_cif_with_biopython(cif: PathLike) -> PathLike:
    """
    Helper function to convert a cif file to a pdb file using biopython.
    """
    from Bio.PDB import MMCIFParser, PDBIO

    if not isinstance(cif, Path):
        cif = Path(cif)
    pdb = cif.with_suffix('.pdb')
    
    parser = MMCIFParser()
    structure = parser.get_structure('protein', str(cif))

    io = PDBIO()
    io.set_structure(structure)
    io.save(str(pdb))

    return pdb

def convert_cif_with_gemmi(cif: PathLike) -> PathLike:
    import gemmi
    structure = gemmi.read_structure(str(cif))
    structure.write(str(cif.with_suffix('.pdb')))

def add_chains(pdb: PathLike,
               first_res: int=1,
               last_res: int=-1) -> PathLike:
    """
    Helper function to add chain IDs to a model.
    """
    import MDAnalysis as mda

    u = mda.Universe(pdb)
    u.add_TopologyAttr('chainID')

    if last_res == -1:
        last_res = len(u.n_residues)

    chain_A = u.select_atoms(f'resid {first_res} to {last_res}')
    chain_A.atoms.chainIDs = 'A'

    if last_res != -1:
        final_res = len(u.n_residues)

        chain_B = u.select_atoms(f'resid {last_res} to {final_res}')
        chain_B.atoms.chainIDs = 'B'

    with mda.Writer(Path(pdb).with_suffix('_withchains.pdb')) as W:
        W.write(u.atoms)



#!/usr/bin/env python
import gc
import MDAnalysis as mda
from MDAnalysis.analysis import align
from MDAnalysis.analysis.contacts import Contacts
from MDAnalysis.analysis.distances import distance_array
from MDAnalysis.analysis.rms import RMSD
from MDAnalysis.analysis.base import AnalysisBase
import numpy as np
import os
from typing import Union

class DeltaCOM(AnalysisBase):
    """
    Computes the change in CoM in a sliding window, saving the delta distances into
    an array. Then loops over the array to determine which if any resident binding
    events have occurred, how long they occurred for and in what order they occurred.
    """
    def __init__(self, sel: mda.AtomGroup, sliding_window: int=1, residence: bool=True,
                 residence_distance_cutoff: float=0.5, min_residence: int=5):
        super().__init__(sel.universe.trajectory)
        self.sel = sel
        self.window = 1
        self.residence = residence
        self.cutoff = residence_distance_cutoff
        self.min_res = min_residence

    def _prepare(self):
        self.coms = np.zeros((len(self._trajectory), 3))
        self.delta_coms = np.zeros((len(self._trajectory)))

    def _single_frame(self):
        fr = self._frame_index
        self.coms[fr, :] = self.sel.center_of_mass()
        self.delta_coms[fr] = self.distance(fr)

    def _conclude(self):
        if self.residence:
            self.residences = []
            length = 1
            for delta in self.delta_coms[1:]:
                if delta <= self.cutoff:
                    length += 1
                else:
                    self.residences.append(length)
                    length = 1

            self.residences = np.array(self.residences)
            self.residences = self.residences[np.where(self.residences >= self.min_res)]

    def distance(self, frame: int):
        """
        Euclidean distance is just L2 norm, use numpy.linalg.norm to
        compute this efficiently, or if its the first frame and we throw
        an IndexError return a NaN.
        """
        try:
            return np.linalg.norm(self.coms[frame, :] - self.coms[frame-self.window, :])
        except IndexError:
            return np.nan


class CoMDist(AnalysisBase):
    def __init__(self, target: mda.AtomGroup, binder: mda.AtomGroup):
        super().__init__(target.universe.trajectory)
        self.target = target
        self.binder = binder

    def _prepare(self):
        self.com_dists = np.zeros((len(self._trajectory)))

    def _single_frame(self):
        pos1 = self.target.center_of_mass()
        pos2 = self.binder.center_of_mass()

        self.com_dists[self._frame_index] = np.linalg.norm(pos1 - pos2)

    def _conclude(self):
        self.rel_dists = self.com_dists - self.com_dists[0]


class RadiusofGyration(AnalysisBase):
    def __init__(self, atomgroup):
        super().__init__(atomgroup.universe.trajectory)
        self.ag = atomgroup
        self.masses = atomgroup.masses
        self.total_mass = np.sum(self.masses)

    def _prepare(self):
        self.results = np.zeros((self.n_frames, 4))

    def _single_frame(self):
        rogs = self.radgyr(self.ag, self.masses, 
                           total_mass=self.total_mass)
        self.results[self._frame_index] = rogs

    def _conclude(self):
        pass

    @staticmethod
    def radgyr(atomgroup, masses, total_mass=None):
        coordinates = atomgroup.positions
        center_of_mass = atomgroup.center_of_mass()

        ri_sq = (coordinates-center_of_mass)**2
        sq = np.sum(ri_sq, axis=1)
        sq_x = np.sum(ri_sq[:,[1,2]], axis=1) # sum over y and z
        sq_y = np.sum(ri_sq[:,[0,2]], axis=1) # sum over x and z
        sq_z = np.sum(ri_sq[:,[0,1]], axis=1) # sum over x and y

        sq_rs = np.array([sq, sq_x, sq_y, sq_z])

        rog_sq = np.sum(masses*sq_rs, axis=1)/total_mass
        return np.sqrt(rog_sq)


class ContactFrequency(AnalysisBase):
    def __init__(self, binder, target, cutoff=5.):
        super().__init__(binder.universe.trajectory)
        self.binder = binder
        self.target = target
        self.cutoff = cutoff

    def _prepare(self):
        self.distance_array = np.zeros((self.n_frames, self.binder.n_atoms, self.target.n_atoms))
        self.contact_probabilities = np.zeros((self.n_frames))

    def _single_frame(self):
        distance_array(self.binder, self.target, 
                       result=self.distance_array[self._frame_index])
        
    def _conclude(self):
        contacts = np.where(self.distance_array < self.cutoff)
        self.contacts = np.unique(np.column_stack(contacts[:2]), axis=0)

        frames, counts = np.unique(self.contacts[:, 0], return_counts=True)
        for (frame, count) in zip(frames, counts):
            self.contact_probabilities[frame] = count

        self.contact_probabilities /= self.binder.n_atoms




try:
    from .interaction_energy import (StaticInteractionEnergy,
                                     DynamicInteractionEnergy)
except ImportError:
    pass

try:
    from .ipSAE import ipSAE
except ImportError:
    pass

try:
    from .sasa import SASA
except ImportError:
    pass

try:
    from .fingerprinter import Fingerprinter
except ImportError:
    pass

from .utils import EmbedEnergyData



import openmm
import MDAnalysis as mda
from numba import njit
import numpy as np
from pathlib import Path
from typing import Union

from ..simulate import Minimizer

OptPath = Union[Path, str, None]
PathLike = Union[Path, str]

@njit
def unravel_index(n1, 
                  n2):
    a, b = np.empty((n1, n2), dtype=np.int32), np.empty((n1, n2), dtype=np.int32)
    for i in range(n1):
        for j in range(n2):
            a[i,j], b[i,j] = i, j
    return a.ravel(),b.ravel()

@njit
def _dist_mat(xyz1, 
              xyz2):
    n1 = xyz1.shape[0]
    n2 = xyz2.shape[0]
    ndim = xyz1.shape[1]
    dist_mat = np.zeros((n1 * n2))
    i, j = unravel_index(n1, n2)
    for k in range(n1 * n2):
        dr = xyz1[i[k]] - xyz2[j[k]]
        for ri in range(ndim):
            dist_mat[k] += np.square(dr[ri])
    return np.sqrt(dist_mat)

@njit
def dist_mat(xyz1, 
             xyz2):
    n1 = xyz1.shape[0]
    n2 = xyz2.shape[0]
    return _dist_mat(xyz1, xyz2).reshape(n1, n2)

@njit
def electrostatic(distance,
                  charge_i, 
                  charge_j):
    """
    Moeen + screening
    Calculate electrostatic energy between two particles.
    Cutoff at 12 Angstrom without switching.

    Parameters:
        distance (float): distance between particles i and j (nm)
        charge_i (float): charge of particle i (e-)
        charge_j (float): charge of particle j (e-)

    Returns:
        energy (float): Electrostatic energy between particles (kJ/mol)
    """
    # conversion factors:
    #     Avogadro = 6.022e23 molecules/mol
    #     e- to Coloumb = 1.602e-19 C/e-
    #     nm to m = 1e-9 m/nm
    #     1/(4\pi\epsilon_0) = 8.988e9 J*m/C^2
    
    solvent_dielectric = 78.5
    # calculate energy
    if distance > 1.:
        energy = 0.
    else:
        r = distance * 1e-9
        r_cutoff = 1. * 1e-9
        k_rf = 1 / (r_cutoff ** 3) * (solvent_dielectric - 1) / (2 * solvent_dielectric + 1)
        c_rf = 1 / r_cutoff * (3 * solvent_dielectric) / (2 * solvent_dielectric + 1)

        outer_term = 8.988e9 * (charge_i * 1.602e-19) * (charge_j * 1.602e-19)
        energy = outer_term * (1 / r + k_rf * r ** 2 - c_rf) * 6.022e23
    return energy / 1000 # J -> kJ

@njit
def electrostatic_sum(distances,
                      charge_is, charge_js):
    """
    Calculate sum of all electrostatic interactions between two
    sets of particles.

    Parameters:
        distances (np.ndarray): distances between particles,
            shape: (len(charge_is),len(charge_js))
        charge_is (np.ndarray): group i charges
        charge_js (np.ndarray): group j charges
    """
    n = distances.shape[0]
    m = distances.shape[1]

    energy = 0.
    for i in range(n):
        for j in range(m):
            energy += electrostatic(distances[i,j],
                                    charge_is[i],
                                    charge_js[j])
    return energy

@njit
def lennard_jones(distance, 
                  sigma_i, sigma_j,
                  epsilon_i, epsilon_j):
    """
    Calculate LJ energy between two particles.
    Cutoff at 12 Angstrom without switching.

    Parameters
    ----------
    distance (float): distance between particles i and j (nm)
    sigma_i (float): sigma parameter for particle i (nm)
    sigma_j (float): sigma parameter for particle j (nm)
    epsilon_i (float): epsilon parameter for particle i (kJ/mol)
    epsilon_j (float): epsilon parameter for particle j (kJ/mol)

    Returns: energy (float): LJ interaction energy (kJ/mol)
    """
    if distance > 1.2:
        energy = 0.
    else:
        # use combination rules to solve for epsilon and sigma
        sigma_ij = 0.5 * (sigma_i + sigma_j)
        epsilon_ij = np.sqrt(epsilon_i * epsilon_j) 
    
        # calculate energy
        sigma_r = sigma_ij / distance
        sigma_r_6 = sigma_r ** 6
        sigma_r_12 = sigma_r_6 ** 2
        energy = 4. * epsilon_ij * (sigma_r_12 - sigma_r_6)
    return energy

@njit
def lennard_jones_sum(distances,
                      sigma_is, sigma_js,
                      epsilon_is, epsilon_js):
    """
    Calculate sum of all LJ interactions between two sets of
    particles.                                       

    Parameters:
        distances (np.ndarray): distances between particles, 
            shape: (len(sigma_is),len(sigma_js))
        sigma_is (np.ndarray): group i sigma parameters
        sigma_js (np.ndarray): group j sigma parameters
        epsilon_is (np.ndarray): group i epsilon parameters
        epsilon_js (np.ndarray): group j epsilon parameters
    """
    n = distances.shape[0]
    m = distances.shape[1]
    energy = 0.
    for i in range(n):
        for j in range(m):
            energy += lennard_jones(distances[i,j], 
                                    sigma_is[i], sigma_js[j],
                                    epsilon_is[i], sigma_js[j])
    return energy

@njit
def fingerprints(xyzs, charges, sigmas, epsilons,
                 target_resmap, binder_inds):
    """
    Calculates electrostatic fingerprint.
    ES energy between each target residue and all binder residues.

    Returns:
        fingerprints: (np.ndarray, np.ndarray), 
            shape=(n_target_residues, n_target_residues)
    """
    n_target_residues = len(target_resmap)
    es_fingerprint = np.zeros((n_target_residues))
    lj_fingerprint = np.zeros((n_target_residues))
    for i in range(n_target_residues):
        dists = dist_mat(xyzs[target_resmap[i]], xyzs[binder_inds])
        es_fingerprint[i] = electrostatic_sum(dists,
                                              charges[target_resmap[i]],
                                              charges[binder_inds])
        lj_fingerprint[i] = lennard_jones_sum(dists,
                                              sigmas[target_resmap[i]],
                                              sigmas[binder_inds],
                                              epsilons[target_resmap[i]],
                                              epsilons[binder_inds])
    return lj_fingerprint, es_fingerprint

class Fingerprinter:
    """
    Calculates interaction energy fingerprint between target and binder chains. 
    
    Inputs:
        pdb_file (str): path to pdb file
        target_resid_range (list[int]): inclusive range of residue indices (1-based) 
            defining target protein
        binder_resid_range (list[int]): inclusive range of residue indices (1-based) 
            defining binder protein
            
    Usage:
        m = Fingerprinter(*args)
        m.run()
        m.save()
    """
    def __init__(self,
                 topology: PathLike,
                 target_selection: str,
                 binder_selection: str | None = None,
                 out_path: OptPath = None,
                 out_name: str | None = None):
        self.topology = Path(topology)
        self.target_selection = target_selection

        if binder_selection is not None:
            self.binder_selection = binder_selection
        else:
            self.binder_selection = f'not {target_selection}'

        if out_path is None:
            path = self.topology.parent
        else:
            path = Path(out_path)

        if out_name is None:
            self.out = path / 'fingerprint.npz'
        else:
            self.out = path / out_name

    def assign_nonbonded_params(self) -> None:
        # build openmm system
        builder = Minimizer(self.topology)
        system = builder.load_files()

        # extract NB params
        nonbonded = [f for f in system.getForces() if isinstance(f, openmm.NonbondedForce)][0]
        self.epsilons = np.zeros((system.getNumParticles()))
        self.sigmas = np.zeros((system.getNumParticles()))
        self.charges = np.zeros((system.getNumParticles()))
        for ind in range(system.getNumParticles()):
            charge, sigma, epsilon = nonbonded.getParticleParameters(ind)
            self.charges[ind] = charge / charge.unit # elementary charge
            self.sigmas[ind] = sigma / sigma.unit # nm
            self.epsilons[ind] = epsilon / epsilon.unit # kJ/mol

    def load_pdb(self) -> None:
        # load with MDAnalysis
        if self.topology.suffix == '.pdb':
            self.u = mda.Universe(self.topology)
        else:
            if self.topology.with_suffix('.inpcrd').exists():
                coordinates = self.topology.with_suffix('.inpcrd')
            else:
                coordinates = self.topology.with_suffix('.rst7')

            self.u = mda.Universe(self.topology, coordinates)

    def assign_residue_mapping(self) -> None:
        # map each residue index (1-based) to corresponding atom indices
        target = self.u.select_atoms(self.target_selection)
        self.target_resmap = [residue.atoms.ix for residue in target.residues]
        self.target_inds = np.concatenate(self.target_resmap)


        binder = self.u.select_atoms(self.binder_selection)
        self.binder_resmap = [residue.atoms.ix for residue in binder.residues]
        self.binder_inds = np.concatenate(self.binder_resmap)

    def iterate_frames(self) -> None:
        self.target_fingerprint = np.zeros((
            len(self.u.trajectory), len(self.target_resmap), 2
        ))

        self.binder_fingerprint = np.zeros((
            len(self.u.trajectory), len(self.binder_resmap), 2
        ))

        for i, ts in enumerate(self.u.trajectory):
            self.calculate_fingerprints(i)

    def calculate_fingerprints(self,
                               frame_index: int) -> None:
        positions = self.u.atoms.positions * .1 # convert to nm
        self.target_fingerprint[frame_index] = np.vstack(
            fingerprints(
                positions,
                self.charges,
                self.sigmas, self.epsilons,
                self.target_resmap, self.binder_inds
            )
        ).T

        self.binder_fingerprint[frame_index] = np.vstack(
            fingerprints(
                positions,
                self.charges,
                self.sigmas, self.epsilons,
                self.binder_resmap, self.target_inds
            )
        ).T
    
    def run(self) -> None:
        self.assign_nonbonded_params()
        self.load_pdb()
        self.assign_residue_mapping()
        self.iterate_frames()

    def save(self) -> None:
        np.savez(self.out, 
                 target=self.target_fingerprint, 
                 binder=self.binder_fingerprint)
